(*----------------------------------------------------------------------------*
 *  DirectX 9 SDK Summer 2003 C++ common framework adaptation for Delphi      *
 *  by Alexey Barkovoy (e-mail: clootie@ixbt.com)                             *
 *                                                                            *
 *  Desc: Direct3D part of framework.                                         *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 22-Jan-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: D3DUtil.h
//
// Desc: Helper functions and typing shortcuts for Direct3D programming.
//
// Copyright (c) Microsoft Corporation. All rights reserved
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// File: D3DUtil.cpp
//
// Desc: Shortcut macros and functions for using DX objects
//
// Copyright (c) Microsoft Corporation. All rights reserved
//-----------------------------------------------------------------------------
unit D3DUtil;

{$I DirectX.inc}

interface

uses
  Windows, Messages, Direct3D9, D3DX9, DXUtil;

//Clootie: WARNING!!! This function was removed in Summer 2003 version
//-----------------------------------------------------------------------------
// Name: D3DUtil_InitMaterial()
// Desc: Initializes a D3DMATERIAL9 structure, setting the diffuse and ambient
//       colors. It does not set emissive or specular colors.
//-----------------------------------------------------------------------------
procedure D3DUtil_InitMaterial(var mtrl: TD3DMaterial9;
  r: Single = 0.0;
  g: Single = 0.0;
  b: Single = 0.0;
  a: Single = 1.0);{$IFDEF COMPILER6_UP} deprecated;{$ENDIF}


//Clootie: WARNING!!! This function was removed in Summer 2003 version
//-----------------------------------------------------------------------------
// Name: D3DUtil_InitLight()
// Desc: Initializes a D3DLIGHT structure, setting the light position. The
//       diffuse color is set to white, specular and ambient left as black.
//-----------------------------------------------------------------------------
procedure D3DUtil_InitLight(var light: TD3DLight9; ltType: TD3DLightType;
  x: Single = 0.0;
  y: Single = 0.0;
  z: Single = 0.0);{$IFDEF COMPILER6_UP} deprecated;{$ENDIF}


//Clootie: WARNING!!! This function was removed in Summer 2003 version
//-----------------------------------------------------------------------------
// Name: D3DUtil_CreateTexture()
// Desc: Helper function to create a texture. It checks the root path first,
//       then tries the DXSDK media path (as specified in the system registry).
//-----------------------------------------------------------------------------
function D3DUtil_CreateTexture(pd3dDevice: IDirect3DDevice9; strTexture: PChar;
  var ppTexture: IDirect3DTexture9;
  d3dFormat: TD3DFormat = D3DFMT_UNKNOWN): HRESULT;{$IFDEF COMPILER6_UP} deprecated;{$ENDIF}




//-----------------------------------------------------------------------------
// Name: D3DUtil_GetCubeMapViewMatrix()
// Desc: Returns a view matrix for rendering to a face of a cubemap.
//-----------------------------------------------------------------------------
function D3DUtil_GetCubeMapViewMatrix(dwFace: TD3DCubemapFaces): TD3DXMatrix;


//-----------------------------------------------------------------------------
// Name: D3DUtil_GetRotationFromCursor()
// Desc: Returns a quaternion for the rotation implied by the window's cursor
//       position.
//-----------------------------------------------------------------------------
function D3DUtil_GetRotationFromCursor(hWnd_: HWND;
  fTrackBallRadius: Single = 1.0): TD3DXQuaternion;


//-----------------------------------------------------------------------------
// Name: D3DUtil_SetDeviceCursor
// Desc: Builds and sets a cursor for the D3D device based on hCursor.
//-----------------------------------------------------------------------------
function D3DUtil_SetDeviceCursor(pd3dDevice: IDirect3DDevice9; hCursor: HCURSOR;
  bAddWatermark: BOOL): HRESULT;


//-----------------------------------------------------------------------------
// Name: D3DUtil_D3DFormatToString
// Desc: Returns the string for the given D3DFORMAT.
//       bWithPrefix determines whether the string should include the "D3DFMT_"
//-----------------------------------------------------------------------------
function D3DUtil_D3DFormatToString(format: TD3DFormat; bWithPrefix: Boolean = True): PChar;


type
  //-----------------------------------------------------------------------------
  // Name: class CD3DArcBall
  // Desc:
  //-----------------------------------------------------------------------------
  CD3DArcBall = class
  protected
    m_mRotation: TD3DXMatrixA16;         // Matrix for arc ball's orientation
    m_mTranslation: TD3DXMatrixA16;      // Matrix for arc ball's position
    m_mTranslationDelta: TD3DXMatrixA16; // Matrix for arc ball's position

    m_nWidth: Integer;                   // arc ball's window width
    m_nHeight: Integer;                  // arc ball's window height
    m_vCenter: TD3DXVector2;             // center of arc ball
    m_fRadius: Single;                   // arc ball's radius in screen coords
    m_fRadiusTranslation: Single;        // arc ball's radius for translating the target

    m_qDown: TD3DXQuaternion;            // Quaternion before button down
    m_qNow: TD3DXQuaternion;             // Composite quaternion for current drag
    m_bDrag: Boolean;                    // Whether user is dragging arc ball

    m_ptLastMouse: TPoint;               // position of last mouse point
    m_vDownPt: TD3DXVector3;             // starting point of rotation arc
    m_vCurrentPt: TD3DXVector3;          // current point of rotation arc

    function QuatFromBallPoints(const vFrom, vTo: TD3DXVector3): TD3DXQuaternion;
    function ScreenToVector(fScreenPtX, fScreenPtY: Single): TD3DXVector3;
    
  public
    constructor Create;

    // Functions to change behavior
    procedure Reset;
    procedure SetTranslationRadius(fRadiusTranslation: Single); { m_fRadiusTranslation = fRadiusTranslation; }
    procedure SetWindow(nWidth: Integer; nHeight: Integer; fRadius: Single = 0.9); { m_nWidth = nWidth; m_nHeight = nHeight; m_fRadius = fRadius; m_vCenter = D3DXVECTOR2(m_nWidth/2.0f,m_nHeight/2.0f); }

    // Call these from client and use GetRotationMatrix() to read new rotation matrix
    procedure OnBegin(nX, nY: Integer);  // start the rotation (pass current mouse position)
    procedure OnMove(nX, nY: Integer);   // continue the rotation (pass current mouse position)
    procedure OnEnd;                     // end the rotation

    // Or call this to automatically handle left, middle, right buttons
    function HandleMessages(hWnd: HWND; uMsg: Cardinal; wParam: WPARAM; lParam: LPARAM): LRESULT; overload;
    procedure HandleMessages(var Msg: TMsg; var Handled: Boolean); overload;

    // Functions to get/set state
    function GetRotationMatrix: PD3DXMatrix; { return D3DXMatrixRotationQuaternion(&m_mRotation, &m_qNow); }
    function GetTranslationMatrix: PD3DXMatrix;      { return &m_mTranslation; }
    function GetTranslationDeltaMatrix: PD3DXMatrix; { return &m_mTranslationDelta; }
    function IsBeingDragged: Boolean;                { return m_bDrag; }
    function GetQuatNow: TD3DXQuaternion;            { return m_qNow; }
    procedure SetQuatNow(q: TD3DXQuaternion);        { m_qNow = q; }
  end;




//-----------------------------------------------------------------------------
// Name: enum D3DUtil_CameraKeys
// Desc: used by CCamera to map WM_KEYDOWN keys
//-----------------------------------------------------------------------------
  TD3DUtil_CameraKeys = (
    CAM_STRAFE_LEFT{ = 0},
    CAM_STRAFE_RIGHT,
    CAM_MOVE_FORWARD,
    CAM_MOVE_BACKWARD,
    CAM_MOVE_UP,
    CAM_MOVE_DOWN,
    CAM_RESET,
    CAM_MAX_KEYS,
    CAM_UNKNOWN{$IFDEF COMPILER6_UP} = $FF{$ENDIF}
  );

const
  KEY_WAS_DOWN_MASK = $80;
  KEY_IS_DOWN_MASK  = $01;

  MOUSE_LEFT_BUTTON   = $01;
  MOUSE_MIDDLE_BUTTON = $02;
  MOUSE_RIGHT_BUTTON  = $04;
  MOUSE_WHEEL         = $08;

type
  //-----------------------------------------------------------------------------
  // Name: class CBaseCamera
  // Desc: Simple base camera class that moves and rotates.  The base class
  //       records mouse and keyboard input for use by a derieved class, and
  //       keeps common state.
  //-----------------------------------------------------------------------------
  CBaseCamera = class
  protected
    m_mView:            TD3DXMatrix;              // View matrix
    m_mProj:            TD3DXMatrix;              // Projection matrix

    m_aKeys: array[CAM_STRAFE_LEFT..CAM_RESET] of Byte; // State of input - KEY_WAS_DOWN_MASK|KEY_IS_DOWN_MASK
    m_ptLastMousePosition: TPoint;        // Last absolute postion of mouse cursor
    m_bMouseLButtonDown: Boolean;         // True if left button is down
    m_bMouseMButtonDown: Boolean;         // True if middle button is down
    m_bMouseRButtonDown: Boolean;         // True if right button is down
    m_nCurrentButtonMask: Integer;        // mask of which buttons are down
    m_nMouseWheelDelta: Integer;          // Amount of middle wheel scroll (+/-)
    m_vMouseDelta: TD3DXVector2;          // Mouse relative delta smoothed over a few frames
    m_fFramesToSmoothMouseData: Single;   // Number of frames to smooth mouse data over

    m_vDefaultEye: TD3DXVector3;          // Default camera eye position
    m_vDefaultLookAt: TD3DXVector3;       // Default LookAt position
    m_vEye: TD3DXVector3;                 // Camera eye position
    m_vLookAt: TD3DXVector3;              // LookAt position
    m_fCameraYawAngle: Single;            // Yaw angle of camera
    m_fCameraPitchAngle: Single;          // Pitch angle of camera

    m_vVelocity: TD3DXVector3;            // Velocity of camera
    m_bMovementDrag: Boolean;             // If true, then camera movement will slow to a stop otherwise movement is instant
    m_vVelocityDrag: TD3DXVector3;        // Velocity drag force
    m_fDragTimer: Single;                 // Countdown timer to apply drag
    m_fTotalDragTimeToZero: Single;       // Time it takes for velocity to go from full to 0
    m_vRotVelocity: TD3DXVector2;         // Velocity of camera

    m_fFOV: Single;                       // Field of view
    m_fAspect: Single;                    // Aspect ratio
    m_fNearPlane: Single;                 // Near plane
    m_fFarPlane: Single;                  // Far plane

    m_fRotationScaler: Single;            // Scaler for rotation
    m_fMoveScaler: Single;                // Scaler for movement

    m_bInvertPitch: Boolean;              // Invert the pitch axis
    m_bEnablePositionMovement: Boolean;   // If true, then the user can translate the camera/model
    m_bEnableYAxisMovement: Boolean;      // If true, then camera can move in the y-axis

    m_bClipToBoundary: Boolean;           // If true, then the camera will be clipped to the boundary
    m_vMinBoundary: TD3DXVector3;         // Min point in clip boundary
    m_vMaxBoundary: TD3DXVector3;         // Max point in clip boundary

    m_bResetCursorAfterMove: Boolean;     // If true, the class will reset the cursor position so that the cursor always has space to move

    // Functions to map a WM_KEYDOWN key to a D3DUtil_CameraKeys enum
    function MapKey(nKey: LongWord): TD3DUtil_CameraKeys; virtual;
    function IsKeyDown(key: Byte): Boolean;  { return( (key & KEY_IS_DOWN_MASK) == KEY_IS_DOWN_MASK ); }
    function WasKeyDown(key: Byte): Boolean; { return( (key & KEY_WAS_DOWN_MASK) == KEY_WAS_DOWN_MASK ); }

    procedure ConstrainToBoundary(var pV: TD3DXVector3);
    procedure UpdateMouseDelta(fElapsedTime: Single);
    procedure UpdateVelocity(fElapsedTime: Single);

  public
    constructor Create;

    // Call these from client and use Get*Matrix() to read new matrices
    function HandleMessages(hWnd: HWND; uMsg: Cardinal; wParam: WPARAM; lParam: LPARAM): LRESULT; virtual;
    procedure FrameMove(fElapsedTime: Single); virtual; abstract;

    // Functions to change camera matrices
    procedure Reset; virtual;
    procedure SetViewParams(const pvEyePt, pvLookatPt: TD3DXVector3);
    procedure SetProjParams(fFOV, fAspect, fNearPlane, fFarPlane: Single);

    // Functions to change behavior
    procedure SetInvertPitch(bInvertPitch: Boolean); { m_bInvertPitch = bInvertPitch; }
    procedure SetDrag(bMovementDrag: Boolean; fTotalDragTimeToZero: Single = 0.25); { m_bMovementDrag = bMovementDrag; m_fTotalDragTimeToZero = fTotalDragTimeToZero; }
    procedure SetEnableYAxisMovement(bEnableYAxisMovement: Boolean); { m_bEnableYAxisMovement = bEnableYAxisMovement; }
    procedure SetEnablePositionMovement(bEnablePositionMovement: Boolean); { m_bEnablePositionMovement = bEnablePositionMovement; }
    procedure SetClipToBoundary(bClipToBoundary: Boolean; const pvMinBoundary, pvMaxBoundary: TD3DXVector3); { m_bClipToBoundary = bClipToBoundary; if( pvMinBoundary ) m_vMinBoundary = *pvMinBoundary; if( pvMaxBoundary ) m_vMaxBoundary = *pvMaxBoundary; }
    procedure SetScalers(fRotationScaler: Single = 0.01; fMoveScaler: Single = 5.0);  { m_fRotationScaler = fRotationScaler; m_fMoveScaler = fMoveScaler; }
    procedure SetNumberOfFramesToSmoothMouseData(nFrames: Integer); { if( nFrames > 0 ) m_fFramesToSmoothMouseData = (float)nFrames; }
    procedure SetResetCursorAfterMove(bResetCursorAfterMove: Boolean); { m_bResetCursorAfterMove = bResetCursorAfterMove; }

    // Functions to get state
    function GetViewMatrix: PD3DXMatrix;      { return &m_mView; }
    function GetProjMatrix: PD3DXMatrix;      { return &m_mProj; }
    function IsBeingDragged: Boolean;         { return (m_bMouseLButtonDown || m_bMouseMButtonDown || m_bMouseRButtonDown); }
    property IsMouseLButtonDown: Boolean      read m_bMouseLButtonDown;
    property IsMouseMButtonDown: Boolean      read m_bMouseMButtonDown;
    property IsMouseRButtonDown: Boolean      read m_bMouseRButtonDown;
  end;



  //-----------------------------------------------------------------------------
  // Name: class CFirstPersonCamera
  // Desc: Simple first person camera class that moves and rotates.
  //       It allows yaw and pitch but not roll.  It uses WM_KEYDOWN and
  //       GetCursorPos() to respond to keyboard and mouse input and updates the
  //       view matrix based on input.
  //-----------------------------------------------------------------------------
  CFirstPersonCamera = class(CBaseCamera)
  protected
    m_mCameraWorld: TD3DXMatrix; // World matrix of the camera (inverse of the view matrix)
  public
    constructor Create;

    // Call these from client and use Get*Matrix() to read new matrices
    procedure FrameMove(fElapsedTime: Single); override;

    // Functions to get state
    function GetWorldRight: TD3DXVector3; { return D3DXVECTOR3( m_mCameraWorld._11, m_mCameraWorld._12, m_mCameraWorld._13 ); }
    function GetWorldUp: TD3DXVector3;    { return D3DXVECTOR3( m_mCameraWorld._21, m_mCameraWorld._22, m_mCameraWorld._23 ); }
    function GetWorldAhead: TD3DXVector3; { return D3DXVECTOR3( m_mCameraWorld._31, m_mCameraWorld._32, m_mCameraWorld._33 ); }
    function GetEyePt: TD3DXVector3;      { return D3DXVECTOR3( m_mCameraWorld._41, m_mCameraWorld._42, m_mCameraWorld._43 ); }
  end;

  //-----------------------------------------------------------------------------
  // Name: class CModelViewerCamera
  // Desc:
  //-----------------------------------------------------------------------------
  CModelViewerCamera = class(CBaseCamera)
  protected
    m_WorldArcBall: CD3DArcBall;
    m_ViewArcBall: CD3DArcBall;
    m_vModelCenter: TD3DXVector3;
    m_mModelLastRot: TD3DXMatrix;        // Last arcball rotation matrix for model
    m_mModelRot: TD3DXMatrix;            // Rotation matrix of model
    m_mWorld: TD3DXMatrix;               // World matrix of model

    m_nRotateModelButtonMask: Integer;
    m_nZoomButtonMask: Integer;
    m_nRotateCameraButtonMask: Integer;

    m_bLimitPitch: Boolean;
    m_fRadius: Single;              // Distance from the camera to model
    m_fDefaultRadius: Single;       // Distance from the camera to model
    m_fMinRadius: Single;           // Min radius
    m_fMaxRadius: Single;           // Max radius
  public
    constructor Create;
    destructor Destroy; override;

    // Call these from client and use Get*Matrix() to read new matrices
    function HandleMessages(hWnd: HWND; uMsg: Cardinal; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
    procedure FrameMove(fElapsedTime: Single); override;

    // Functions to change behavior
    procedure Reset; override;
    procedure SetButtonMasks(nRotateModelButtonMask: Integer = MOUSE_LEFT_BUTTON; nZoomButtonMask: Integer = MOUSE_WHEEL; nRotateCameraButtonMask: Integer = MOUSE_RIGHT_BUTTON); { m_nRotateModelButtonMask = nRotateModelButtonMask, m_nZoomButtonMask = nZoomButtonMask; m_nRotateCameraButtonMask = nRotateCameraButtonMask; }
    procedure SetWindow(nWidth, nHeight: Integer; fArcballRadius: Single = 0.9); { m_WorldArcBall.SetWindow( nWidth, nHeight, fArcballRadius ); m_ViewArcBall.SetWindow( nWidth, nHeight, fArcballRadius ); }
    procedure SetRadius(fDefaultRadius: Single = 5.0; fMinRadius: Single = 1.0; fMaxRadius: Single = 3.402823466e+38{FLT_MAX}); { m_fDefaultRadius = m_fRadius = fDefaultRadius; m_fMinRadius = fMinRadius; m_fMaxRadius = fMaxRadius; }
    procedure SetModelCenter(vModelCenter: TD3DXVector3); { m_vModelCenter = vModelCenter; }
    procedure SetLimitPitch(bLimitPitch: Boolean); { m_bLimitPitch = bLimitPitch; }
    procedure SetViewQuat(q: TD3DXQuaternion); { m_ViewArcBall.SetQuatNow( q ); }
    procedure SetWorldQuat(q: TD3DXQuaternion); { m_WorldArcBall.SetQuatNow( q ); }

    // Functions to get state
    function GetWorldMatrix: PD3DXMatrix; { return &m_mWorld; }
  end;

implementation

uses Math, SysUtils;


//Clootie: WARNING!!! This function was removed in Summer 2003 version
//-----------------------------------------------------------------------------
// Name: D3DUtil_InitMaterial()
// Desc: Initializes a D3DMATERIAL9 structure, setting the diffuse and ambient
//       colors. It does not set emissive or specular colors.
//-----------------------------------------------------------------------------
procedure D3DUtil_InitMaterial(var mtrl: TD3DMaterial9; r: Single; g: Single;
                                                        b: Single; a: Single);
begin
  ZeroMemory(@mtrl, SizeOf(TD3DMaterial9));
  mtrl.Diffuse.r := r;
  mtrl.Ambient.r := r;
  mtrl.Diffuse.g := g;
  mtrl.Ambient.g := g;
  mtrl.Diffuse.b := b;
  mtrl.Ambient.b := b;
  mtrl.Diffuse.a := a;
  mtrl.Ambient.a := a;
end;



//Clootie: WARNING!!! This function was removed in Summer 2003 version
//-----------------------------------------------------------------------------
// Name: D3DUtil_InitLight()
// Desc: Initializes a D3DLIGHT structure, setting the light position. The
//       diffuse color is set to white; specular and ambient are left as black.
//-----------------------------------------------------------------------------
procedure D3DUtil_InitLight(var light: TD3DLight9; ltType: TD3DLightType;
  x: Single; y: Single; z: Single);
var
  vecLightDirUnnormalized: TD3DXVector3;
begin
  vecLightDirUnnormalized:= D3DXVector3(x, y, z);
  ZeroMemory(@light, SizeOf(TD3DLight9));
  light._Type       := ltType;
  light.Diffuse.r   := 1.0;
  light.Diffuse.g   := 1.0;
  light.Diffuse.b   := 1.0;
  D3DXVec3Normalize(light.Direction, vecLightDirUnnormalized);
  light.Position.x  := x;
  light.Position.y  := y;
  light.Position.z  := z;
  light.Range       := 1000.0;
end;



//Clootie: WARNING!!! This function was removed in Summer 2003 version
//-----------------------------------------------------------------------------
// Name: D3DUtil_CreateTexture()
// Desc: Helper function to create a texture. It checks the root path first,
//       then tries the DXSDK media path (as specified in the system registry).
//-----------------------------------------------------------------------------
function D3DUtil_CreateTexture(pd3dDevice: IDirect3DDevice9; strTexture: PChar;
  var ppTexture: IDirect3DTexture9; d3dFormat: TD3DFormat): HRESULT;
var
  strPath: array[0..MAX_PATH-1] of Char;
begin
  // Get the path to the texture
  Result:= DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), strTexture);
  if FAILED(Result) then Exit;

  // Create the texture using D3DX
  Result:= D3DXCreateTextureFromFileEx(pd3dDevice, strPath,
              D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, d3dFormat,
              D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
              D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, ppTexture);
end;



//-----------------------------------------------------------------------------
// Name: D3DUtil_GetCubeMapViewMatrix()
// Desc: Returns a view matrix for rendering to a face of a cubemap.
//-----------------------------------------------------------------------------
function D3DUtil_GetCubeMapViewMatrix(dwFace: TD3DCubemapFaces): TD3DXMatrix;
var
  vEyePt: TD3DXVector3;
  vLookDir: TD3DXVector3;
  vUpDir: TD3DXVector3;
  mView: TD3DXMatrixA16;
begin
  vEyePt:= D3DXVector3(0.0, 0.0, 0.0);

  case dwFace of
    D3DCUBEMAP_FACE_POSITIVE_X:
    begin
      vLookDir := D3DXVector3( 1.0, 0.0, 0.0 );
      vUpDir   := D3DXVector3( 0.0, 1.0, 0.0 );
    end;
    D3DCUBEMAP_FACE_NEGATIVE_X:
    begin
      vLookDir := D3DXVector3(-1.0, 0.0, 0.0 );
      vUpDir   := D3DXVector3( 0.0, 1.0, 0.0 );
    end;
    D3DCUBEMAP_FACE_POSITIVE_Y:
    begin
      vLookDir := D3DXVector3( 0.0, 1.0, 0.0 );
      vUpDir   := D3DXVector3( 0.0, 0.0,-1.0 );
    end;
    D3DCUBEMAP_FACE_NEGATIVE_Y:
    begin
      vLookDir := D3DXVector3( 0.0,-1.0, 0.0 );
      vUpDir   := D3DXVector3( 0.0, 0.0, 1.0 );
    end;
    D3DCUBEMAP_FACE_POSITIVE_Z:
    begin
      vLookDir := D3DXVector3( 0.0, 0.0, 1.0 );
      vUpDir   := D3DXVector3( 0.0, 1.0, 0.0 );
    end;
    D3DCUBEMAP_FACE_NEGATIVE_Z:
    begin
      vLookDir := D3DXVector3( 0.0, 0.0,-1.0 );
      vUpDir   := D3DXVector3( 0.0, 1.0, 0.0 );
    end;
  end;

  // Set the view transform for this cubemap surface
  D3DXMatrixLookAtLH(mView, vEyePt, vLookDir, vUpDir);
  Result:= mView;
end;



//-----------------------------------------------------------------------------
// Name: D3DUtil_GetRotationFromCursor()
// Desc: Returns a quaternion for the rotation implied by the window's cursor
//       position.
//-----------------------------------------------------------------------------
function D3DUtil_GetRotationFromCursor(hWnd_: HWND; fTrackBallRadius: Single): TD3DXQuaternion;
var
  pt: TPoint;
  rc: TRect;
  sx, sy, sz, d2, t, fAngle: Single;
  p1: TD3DXVector3;
  p2: TD3DXVector3;
  vAxis: TD3DXVector3;
  quat: TD3DXQuaternion;
  v: TD3DXVector3; // temporary vector
  vecDiff: TD3DXVector3;
begin
  GetCursorPos(pt);
  GetClientRect(hWnd_, rc);
  ScreenToClient(hWnd_, pt);
  sx := ((2.0 * pt.x) / (rc.right-rc.left)) - 1;
  sy := ((2.0 * pt.y) / (rc.bottom-rc.top)) - 1;

  if (sx = 0.0) and (sy = 0.0) then
  begin
    Result:= D3DXQuaternion(0.0, 0.0, 0.0, 1.0);
    Exit;
  end;

  d2:= Sqrt(sx*sx + sy*sy);

  if (d2 < fTrackBallRadius * 0.70710678118654752440) then // Inside sphere
    sz := Sqrt(fTrackBallRadius*fTrackBallRadius - d2*d2)
  else                                                     // On hyperbola
    sz := (fTrackBallRadius*fTrackBallRadius) / (2.0*d2);

  // Get two points on trackball's sphere
  p1:= D3DXVector3(sx, sy, sz);
  p2:= D3DXVector3(0.0, 0.0, fTrackBallRadius);

  // Get axis of rotation, which is cross product of p1 and p2
  D3DXVec3Cross(vAxis, p1, p2);

  // Calculate angle for the rotation about that axis
  D3DXVec3Subtract(vecDiff, p2, p1);
  // FLOAT t = D3DXVec3Length( &vecDiff ) / ( 2.0f*fTrackBallRadius );
  t:= D3DXVec3Length(D3DXVec3Scale(v, vecDiff, 1 / (2.0*fTrackBallRadius))^);
  if (t > +1.0) then t := +1.0;
  if (t < -1.0) then t := -1.0;
  fAngle := 2.0 * ArcSin(t);

  // Convert axis to quaternion
  D3DXQuaternionRotationAxis(quat, vAxis, fAngle);
  Result:= quat;
end;



//-----------------------------------------------------------------------------
// Name: D3DUtil_SetDeviceCursor
// Desc: Gives the D3D device a cursor with image and hotspot from hCursor.
//-----------------------------------------------------------------------------
function D3DUtil_SetDeviceCursor(pd3dDevice: IDirect3DDevice9; hCursor: HCURSOR;
  bAddWatermark: BOOL): HRESULT;
const
  wMask: array [0..4] of Word = ($ccc0, $a2a0, $a4a0, $a2a0, $ccc0);
label
  End_;
type
  PACOLORREF = ^ACOLORREF;
  ACOLORREF = array[0..0] of COLORREF;
type
  pImg = ^img;
  img = array[0..16000] of DWORD;
var
  hr: HRESULT;
  iconinfo_: TIconInfo;
  bBWCursor: BOOL;
  pCursorSurface: IDirect3DSurface9;
  hdcColor: HDC;
  hdcMask: HDC;
  hdcScreen: HDC;
  bm: TBitmap;
  dwWidth: DWORD;
  dwHeightSrc: DWORD;
  dwHeightDest: DWORD;
  crColor: COLORREF;
  crMask: COLORREF;
  x,y: Cardinal;
  bmi: TBitmapInfo;
  pcrArrayColor: PACOLORREF;
  pcrArrayMask: PACOLORREF;
  pBitmap: pImg;
  hgdiobjOld: HGDIOBJ;
  lr: TD3DLockedRect;
begin
  hr := E_FAIL;
  pCursorSurface := nil;
  hdcColor := 0;
  hdcMask := 0;
  hdcScreen := 0;
  pcrArrayColor := nil;
  pcrArrayMask := nil;

  ZeroMemory(@iconinfo_, SizeOf(TIconInfo));
  if not GetIconInfo(hCursor, iconinfo_) then
    goto End_;

  if (0 = GetObject(iconinfo_.hbmMask, SizeOf(TBitmap), @bm)) then
    goto End_;
  dwWidth := bm.bmWidth;
  dwHeightSrc := bm.bmHeight;

  if (iconinfo_.hbmColor = 0) then
  begin
    bBWCursor := TRUE;
    dwHeightDest := dwHeightSrc div 2;
  end else
  begin
    bBWCursor := FALSE;
    dwHeightDest := dwHeightSrc;
  end;

  // Create a surface for the fullscreen cursor
  hr:= pd3dDevice.CreateOffscreenPlainSurface(dwWidth, dwHeightDest,
          D3DFMT_A8R8G8B8, D3DPOOL_SCRATCH, pCursorSurface, nil);
  if FAILED(hr) then
    goto End_;

  // pcrArrayMask = new DWORD[dwWidth * dwHeightSrc];
  GetMem(pcrArrayMask, SizeOf(DWORD)*(dwWidth * dwHeightSrc));

  ZeroMemory(@bmi, sizeof(bmi));
  bmi.bmiHeader.biSize := sizeof(bmi.bmiHeader);
  bmi.bmiHeader.biWidth := dwWidth;
  bmi.bmiHeader.biHeight := dwHeightSrc;
  bmi.bmiHeader.biPlanes := 1;
  bmi.bmiHeader.biBitCount := 32;
  bmi.bmiHeader.biCompression := BI_RGB;

  hdcScreen := GetDC(0);
  hdcMask := CreateCompatibleDC(hdcScreen);
  if (hdcMask = 0) then
  begin
    hr := E_FAIL;
    goto End_;
  end;
  hgdiobjOld := SelectObject(hdcMask, iconinfo_.hbmMask);
  GetDIBits(hdcMask, iconinfo_.hbmMask, 0, dwHeightSrc, pcrArrayMask, bmi,
    DIB_RGB_COLORS);
  SelectObject(hdcMask, hgdiobjOld);

  if (not bBWCursor) then
  begin
    // pcrArrayColor = new DWORD[dwWidth * dwHeightDest];
    GetMem(pcrArrayColor, SizeOf(DWORD)*(dwWidth * dwHeightDest));
    hdcColor := CreateCompatibleDC(hdcScreen);
    if (hdcColor = 0) then
    begin
      hr := E_FAIL;
      goto End_;
    end;
    SelectObject(hdcColor, iconinfo_.hbmColor);
    GetDIBits(hdcColor, iconinfo_.hbmColor, 0, dwHeightDest, pcrArrayColor, bmi,
      DIB_RGB_COLORS);
  end;

  // Transfer cursor image into the surface
  pCursorSurface.LockRect(lr, nil, 0);
  pBitmap:= lr.pBits;
  for y:= 0 to dwHeightDest - 1 do
  begin
    for x:= 0 to dwWidth - 1 do
    begin
      if bBWCursor then
      begin
        crColor:= pcrArrayMask^[dwWidth*(dwHeightDest-1-y) + x];
        crMask:= pcrArrayMask^[dwWidth*(dwHeightSrc-1-y) + x];
      end else
      begin
        crColor:= pcrArrayColor^[dwWidth*(dwHeightDest-1-y) + x];
        crMask:= pcrArrayMask^[dwWidth*(dwHeightDest-1-y) + x];
      end;
      if (crMask = 0) then
        pBitmap^[dwWidth*y + x]:= $ff000000 or crColor
      else
        pBitmap^[dwWidth*y + x]:= $00000000;

      // It may be helpful to make the D3D cursor look slightly
      // different from the Windows cursor so you can distinguish
      // between the two when developing/testing code.  When
      // bAddWatermark is TRUE, the following code adds some
      // small grey "D3D" characters to the upper-left corner of
      // the D3D cursor image.

      //if( bAddWatermark && x < 12 && y < 5 )
      if bAddWatermark and (x < 12) and (y < 5) then
      begin
          // 11.. 11.. 11.. .... CCC0
          // 1.1. ..1. 1.1. .... A2A0
          // 1.1. .1.. 1.1. .... A4A0
          // 1.1. ..1. 1.1. .... A2A0
          // 11.. 11.. 11.. .... CCC0

          // if( wMask[y] & (1 << (15 - x)) )
          if (wMask[y] and (1 shl (15 - x)) <> 0) then
          begin
            pBitmap^[dwWidth*y + x]:= pBitmap^[dwWidth*y + x] or $ff808080;
          end;
      end;
    end;
  end;
  pCursorSurface.UnlockRect;

  // Set the device cursor
  hr := pd3dDevice.SetCursorProperties(iconinfo_.xHotspot,
      iconinfo_.yHotspot, pCursorSurface);
  if FAILED(hr) then
    goto End_;

  hr := S_OK;

End_:
  if (iconinfo_.hbmMask <> 0)  then DeleteObject(iconinfo_.hbmMask);
  if (iconinfo_.hbmColor <> 0) then DeleteObject(iconinfo_.hbmColor);
  if (hdcScreen <> 0)          then ReleaseDC(0, hdcScreen);
  if (hdcColor <> 0)           then DeleteDC(hdcColor);
  if (hdcMask <> 0)            then DeleteDC(hdcMask);
  // SAFE_DELETE_ARRAY(pcrArrayColor);
  FreeMem(pcrArrayColor);
  // SAFE_DELETE_ARRAY(pcrArrayMask);
  FreeMem(pcrArrayMask);
  SAFE_RELEASE(pCursorSurface);
  Result:= hr;
end;



//-----------------------------------------------------------------------------
// Name: D3DFormatToString
// Desc: Returns the string for the given D3DFORMAT.
//-----------------------------------------------------------------------------
function D3DUtil_D3DFormatToString(format: TD3DFormat; bWithPrefix: Boolean = True): PChar;
begin
  case format of
    D3DFMT_UNKNOWN:         Result:= 'D3DFMT_UNKNOWN';
    D3DFMT_R8G8B8:          Result:= 'D3DFMT_R8G8B8';
    D3DFMT_A8R8G8B8:        Result:= 'D3DFMT_A8R8G8B8';
    D3DFMT_X8R8G8B8:        Result:= 'D3DFMT_X8R8G8B8';
    D3DFMT_R5G6B5:          Result:= 'D3DFMT_R5G6B5';
    D3DFMT_X1R5G5B5:        Result:= 'D3DFMT_X1R5G5B5';
    D3DFMT_A1R5G5B5:        Result:= 'D3DFMT_A1R5G5B5';
    D3DFMT_A4R4G4B4:        Result:= 'D3DFMT_A4R4G4B4';
    D3DFMT_R3G3B2:          Result:= 'D3DFMT_R3G3B2';
    D3DFMT_A8:              Result:= 'D3DFMT_A8';
    D3DFMT_A8R3G3B2:        Result:= 'D3DFMT_A8R3G3B2';
    D3DFMT_X4R4G4B4:        Result:= 'D3DFMT_X4R4G4B4';
    D3DFMT_A2B10G10R10:     Result:= 'D3DFMT_A2B10G10R10';
    D3DFMT_A8B8G8R8:        Result:= 'D3DFMT_A8B8G8R8';
    D3DFMT_X8B8G8R8:        Result:= 'D3DFMT_X8B8G8R8';
    D3DFMT_G16R16:          Result:= 'D3DFMT_G16R16';
    D3DFMT_A2R10G10B10:     Result:= 'D3DFMT_A2R10G10B10';
    D3DFMT_A16B16G16R16:    Result:= 'D3DFMT_A16B16G16R16';
    D3DFMT_A8P8:            Result:= 'D3DFMT_A8P8';
    D3DFMT_P8:              Result:= 'D3DFMT_P8';
    D3DFMT_L8:              Result:= 'D3DFMT_L8';
    D3DFMT_A8L8:            Result:= 'D3DFMT_A8L8';
    D3DFMT_A4L4:            Result:= 'D3DFMT_A4L4';
    D3DFMT_V8U8:            Result:= 'D3DFMT_V8U8';
    D3DFMT_L6V5U5:          Result:= 'D3DFMT_L6V5U5';
    D3DFMT_X8L8V8U8:        Result:= 'D3DFMT_X8L8V8U8';
    D3DFMT_Q8W8V8U8:        Result:= 'D3DFMT_Q8W8V8U8';
    D3DFMT_V16U16:          Result:= 'D3DFMT_V16U16';
    D3DFMT_A2W10V10U10:     Result:= 'D3DFMT_A2W10V10U10';
    D3DFMT_UYVY:            Result:= 'D3DFMT_UYVY';
    D3DFMT_YUY2:            Result:= 'D3DFMT_YUY2';
    D3DFMT_DXT1:            Result:= 'D3DFMT_DXT1';
    D3DFMT_DXT2:            Result:= 'D3DFMT_DXT2';
    D3DFMT_DXT3:            Result:= 'D3DFMT_DXT3';
    D3DFMT_DXT4:            Result:= 'D3DFMT_DXT4';
    D3DFMT_DXT5:            Result:= 'D3DFMT_DXT5';
    D3DFMT_D16_LOCKABLE:    Result:= 'D3DFMT_D16_LOCKABLE';
    D3DFMT_D32:             Result:= 'D3DFMT_D32';
    D3DFMT_D15S1:           Result:= 'D3DFMT_D15S1';
    D3DFMT_D24S8:           Result:= 'D3DFMT_D24S8';
    D3DFMT_D24X8:           Result:= 'D3DFMT_D24X8';
    D3DFMT_D24X4S4:         Result:= 'D3DFMT_D24X4S4';
    D3DFMT_D16:             Result:= 'D3DFMT_D16';
    D3DFMT_L16:             Result:= 'D3DFMT_L16';
    D3DFMT_VERTEXDATA:      Result:= 'D3DFMT_VERTEXDATA';
    D3DFMT_INDEX16:         Result:= 'D3DFMT_INDEX16';
    D3DFMT_INDEX32:         Result:= 'D3DFMT_INDEX32';
    D3DFMT_Q16W16V16U16:    Result:= 'D3DFMT_Q16W16V16U16';
    D3DFMT_MULTI2_ARGB8:    Result:= 'D3DFMT_MULTI2_ARGB8';
    D3DFMT_R16F:            Result:= 'D3DFMT_R16F';
    D3DFMT_G16R16F:         Result:= 'D3DFMT_G16R16F';
    D3DFMT_A16B16G16R16F:   Result:= 'D3DFMT_A16B16G16R16F';
    D3DFMT_R32F:            Result:= 'D3DFMT_R32F';
    D3DFMT_G32R32F:         Result:= 'D3DFMT_G32R32F';
    D3DFMT_A32B32G32R32F:   Result:= 'D3DFMT_A32B32G32R32F';
    D3DFMT_CxV8U8:          Result:= 'D3DFMT_CxV8U8';
  else
    Result:= 'Unknown format';
  end;
  if (bWithPrefix or (Pos('D3DFMT_', Result) = 0)) then {nothing}
  else Result:= Result + Length('D3DFMT_');
end;



//-----------------------------------------------------------------------------
// Name: D3DUtil_QuaternionUnitAxisToUnitAxis2
// Desc: Axis to axis quaternion double angle (no normalization)
//       Takes two points on unit sphere an angle THETA apart, returns
//       quaternion that represents a rotation around cross product by 2*THETA.
//-----------------------------------------------------------------------------
{inline} function D3DUtil_QuaternionUnitAxisToUnitAxis2(out pOut: TD3DXQuaternion;
           const pvFrom, pvTo: TD3DXVector3): PD3DXQuaternion; // WINAPI;
var
  vAxis: TD3DXVector3;
begin
  D3DXVec3Cross(vAxis, pvFrom, pvTo);    // proportional to sin(theta)
  pOut.x := vAxis.x;
  pOut.y := vAxis.y;
  pOut.z := vAxis.z;
  pOut.w := D3DXVec3Dot(pvFrom, pvTo);
  Result:= @pOut;
end;



//-----------------------------------------------------------------------------
// Name: D3DUtil_QuaternionAxisToAxis
// Desc: Axis to axis quaternion
//       Takes two points on unit sphere an angle THETA apart, returns
//       quaternion that represents a rotation around cross product by theta.
//-----------------------------------------------------------------------------
{inline} function D3DUtil_QuaternionAxisToAxis(out pOut: TD3DXQuaternion;
           const pvFrom, pvTo: TD3DXVector3): PD3DXQuaternion; // WINAPI;
var
  vA, vB: TD3DXVector3;
  vHalf: TD3DXVector3;
begin
  D3DXVec3Normalize(vA, pvFrom);
  D3DXVec3Normalize(vB, pvTo);
  D3DXVec3Add(vHalf, vA, vB);
  D3DXVec3Normalize(vHalf, vHalf);
  Result:= D3DUtil_QuaternionUnitAxisToUnitAxis2(pOut, vA, vHalf);
end;




//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
constructor CD3DArcBall.Create;
var
  rc: TRect;
begin
  Reset;
  m_vDownPt := D3DXVector3Zero;
  m_vCurrentPt := D3DXVector3Zero;

  GetClientRect(GetForegroundWindow, rc);
  SetWindow(rc.right, rc.bottom);
end;


//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
procedure CD3DArcBall.Reset;
begin
  D3DXQuaternionIdentity(m_qDown);
  D3DXQuaternionIdentity(m_qNow);
  D3DXMatrixIdentity(m_mRotation);
  D3DXMatrixIdentity(m_mTranslation);
  D3DXMatrixIdentity(m_mTranslationDelta);
  m_bDrag := False;
  m_fRadiusTranslation := 1.0;
  m_fRadius := 1.0;
end;


procedure CD3DArcBall.SetTranslationRadius(fRadiusTranslation: Single);
begin
  m_fRadiusTranslation := fRadiusTranslation;
end;


//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
//todo: Move this somewhere below
procedure CD3DArcBall.SetWindow(nWidth: Integer; nHeight: Integer; fRadius: Single = 0.9);
begin
  m_nWidth := nWidth;
  m_nHeight := nHeight;
  m_fRadius := fRadius;
  m_vCenter := D3DXVector2(m_nWidth/2.0, m_nHeight/2.0);
end;


//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
function CD3DArcBall.ScreenToVector(fScreenPtX, fScreenPtY: Single): TD3DXVector3;
var
  x, y, z, mag, scale: Single;
begin
  // Scale to screen
  x   := -(fScreenPtX - m_nWidth/2)  / (m_fRadius*m_nWidth/2);
  y   :=  (fScreenPtY - m_nHeight/2) / (m_fRadius*m_nHeight/2);

  z   := 0.0;
  mag := x*x + y*y;

  if (mag > 1.0) then
  begin
    scale := 1.0 / Sqrt(mag);
    x := x*scale;
    y := y*scale;
  end
  else
    z := Sqrt(1.0 - mag);

  // Return vector
  Result:= D3DXVector3(x, y, z);
end;




//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
function CD3DArcBall.QuatFromBallPoints(const vFrom, vTo: TD3DXVector3): TD3DXQuaternion;
var
  vPart: TD3DXVector3;
  fDot: Single;
begin
  fDot := D3DXVec3Dot(vFrom, vTo);
  D3DXVec3Cross(vPart, vFrom, vTo);

  Result:= D3DXQuaternion(vPart.x, vPart.y, vPart.z, fDot);
end;




//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
procedure CD3DArcBall.OnBegin(nX, nY: Integer);
begin
  m_bDrag := True;
  m_vDownPt := ScreenToVector(nX, nY);
end;




//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
procedure CD3DArcBall.OnMove(nX, nY: Integer);
begin
  if (m_bDrag) then
  begin
    m_vCurrentPt := ScreenToVector(nX, nY);
    D3DXQuaternionMultiply(m_qNow, m_qDown, QuatFromBallPoints(m_vDownPt, m_vCurrentPt));
  end;
end;




//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
procedure CD3DArcBall.OnEnd;
begin
  m_bDrag := False;
  m_qDown := m_qNow;
end;




//-----------------------------------------------------------------------------
// Name: HandleMessages
// Desc:
//-----------------------------------------------------------------------------
function CD3DArcBall.HandleMessages(hWnd: HWND; uMsg: Cardinal; wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  iMouseX: Integer;
  iMouseY: Integer;
  fDeltaX: Single;
  fDeltaY: Single;
// #define GET_X_LPARAM(lp)                        ((int)(short)LOWORD(lp))
// #define GET_Y_LPARAM(lp)                        ((int)(short)HIWORD(lp))
begin
//UNREFERENCED_PARAMETER( hWnd );

  // Current mouse position
  iMouseX := LOWORD(lParam); // GET_X_LPARAM(lParam);
  iMouseY := HIWORD(lParam); // GET_Y_LPARAM(lParam);

  Result:= iTrue;
  case uMsg of
    WM_LBUTTONDOWN:
    begin
      OnBegin(iMouseX, iMouseY);
      Exit;
    end;

    WM_LBUTTONUP:
    begin
      OnEnd();
      Exit;
    end;

    WM_RBUTTONDOWN,
    WM_MBUTTONDOWN:
    begin
      // Store off the position of the cursor when the button is pressed
      m_ptLastMouse.x := iMouseX;
      m_ptLastMouse.y := iMouseY;
      Exit;
    end;

    WM_MOUSEMOVE:
    begin
      if (MK_LBUTTON and wParam <> 0) then
      begin
        OnMove(iMouseX, iMouseY);
      end
      else if (MK_RBUTTON and wParam <> 0) or (MK_MBUTTON and wParam <> 0) then
      begin
        // Normalize based on size of window and bounding sphere radius
        fDeltaX := (m_ptLastMouse.x-iMouseX) * m_fRadiusTranslation / m_nWidth;
        fDeltaY := (m_ptLastMouse.y-iMouseY) * m_fRadiusTranslation / m_nHeight;

        if (wParam and MK_RBUTTON <> 0) then
        begin
          D3DXMatrixTranslation(m_mTranslationDelta, -2*fDeltaX, 2*fDeltaY, 0.0);
          D3DXMatrixMultiply(m_mTranslation, m_mTranslation, m_mTranslationDelta);
        end
        else  // wParam & MK_MBUTTON
        begin
          D3DXMatrixTranslation(m_mTranslationDelta, 0.0, 0.0, 5*fDeltaY);
          D3DXMatrixMultiply(m_mTranslation, m_mTranslation, m_mTranslationDelta);
        end;

        // Store mouse coordinate
        m_ptLastMouse.x := iMouseX;
        m_ptLastMouse.y := iMouseY;
      end;
      Result:= iTrue;
      Exit;
    end;
  end;

  Result:= iFalse;
end;

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
procedure CD3DArcBall.HandleMessages(var Msg: TMsg; var Handled: Boolean);
begin
  Handled := HandleMessages(Msg.hwnd, Msg.message, Msg.wParam, Msg.lParam) <> 0;
end;

function CD3DArcBall.GetRotationMatrix: PD3DXMatrix;
begin
  Result:= D3DXMatrixRotationQuaternion(m_mRotation, m_qNow);
end;

function CD3DArcBall.GetTranslationMatrix: PD3DXMatrix;      { return &m_matTranslation; }
begin
  Result:= @m_mTranslation;
end;

function CD3DArcBall.GetTranslationDeltaMatrix: PD3DXMatrix; { return &m_matTranslationDelta; }
begin
  Result:= @m_mTranslationDelta;
end;

function CD3DArcBall.IsBeingDragged: Boolean;                { return m_bDrag; }
begin
  Result:= m_bDrag;
end;


function CD3DArcBall.GetQuatNow: TD3DXQuaternion;            { return m_qNow; }
begin
  Result:= m_qNow;
end;


procedure CD3DArcBall.SetQuatNow(q: TD3DXQuaternion);        { m_qNow = q; }
begin
  m_qNow := q;
end;


//-----------------------------------------------------------------------------
// Name: CBaseCamera
// Desc: Constructor
//-----------------------------------------------------------------------------
constructor CBaseCamera.Create;
var
  vEyePt: TD3DXVector3;
  vLookatPt: TD3DXVector3;
begin
  ZeroMemory(@m_aKeys, SizeOf(Byte)*DWORD(CAM_MAX_KEYS));

  // Set attributes for the view matrix
  vEyePt    := D3DXVector3(0.0,0.0,0.0);
  vLookatPt := D3DXVector3(0.0,0.0,1.0);

  // Setup the view matrix
  SetViewParams(vEyePt, vLookatPt);

  // Setup the projection matrix
  SetProjParams(D3DX_PI/4, 1.0, 1.0, 1000.0);

  GetCursorPos(m_ptLastMousePosition);
  m_bMouseLButtonDown := False;
  m_bMouseMButtonDown := False;
  m_bMouseRButtonDown := False;
  m_nCurrentButtonMask := 0;
  m_nMouseWheelDelta := 0;

  m_fCameraYawAngle := 0.0;
  m_fCameraPitchAngle := 0.0;

  m_vVelocity     := D3DXVECTOR3(0,0,0);
  m_bMovementDrag := false;
  m_vVelocityDrag := D3DXVECTOR3(0,0,0);
  m_fDragTimer    := 0.0;
  m_fTotalDragTimeToZero := 0.25;
  m_vRotVelocity := D3DXVECTOR2(0,0);

  m_fRotationScaler := 0.01;
  m_fMoveScaler := 5.0;

  m_bInvertPitch := False;
  m_bEnableYAxisMovement := True;
  m_bEnablePositionMovement := True;

  m_vMouseDelta   := D3DXVector2Zero;
  m_fFramesToSmoothMouseData := 2.0;

  m_bClipToBoundary := False;
  m_vMinBoundary := D3DXVector3(-1,-1,-1);
  m_vMaxBoundary := D3DXVector3(1,1,1);

  m_bResetCursorAfterMove := False;
end;


//-----------------------------------------------------------------------------
// Name: SetViewParams
// Desc: Client can call this to change the position and direction of camrea
//-----------------------------------------------------------------------------
procedure CBaseCamera.SetViewParams(const pvEyePt, pvLookatPt: TD3DXVector3);
var
  vUp: TD3DXVector3;
  mInvView: TD3DXMatrix;
  pZBasis: PD3DXVector3;
  fLen: Single;
begin
  m_vEye := pvEyePt;        m_vDefaultEye := pvEyePt;
  m_vLookAt := pvLookatPt;  m_vDefaultLookAt := pvLookatPt;

  // Calc the view matrix
  vUp := D3DXVector3(0,1,0);
  D3DXMatrixLookAtLH(m_mView, pvEyePt, pvLookatPt, vUp);

  D3DXMatrixInverse(mInvView, nil, m_mView);

  // The axis basis vectors and camera position are stored inside the
  // position matrix in the 4 rows of the camera's world matrix.
  // To figuire out the yaw/pitch of the camera, we just need the Z basis vector
  pZBasis := PD3DXVector3(@mInvView._31);

  m_fCameraYawAngle   := ArcTan2(pZBasis^.x, pZBasis^.z);
  fLen := Sqrt(pZBasis^.z*pZBasis^.z + pZBasis^.x*pZBasis^.x);
  m_fCameraPitchAngle := -ArcTan2(pZBasis^.y, fLen);
end;




//-----------------------------------------------------------------------------
// Name: SetProjParams
// Desc: Calculates the projection matrix based on input params
//-----------------------------------------------------------------------------
procedure CBaseCamera.SetProjParams(fFOV, fAspect, fNearPlane, fFarPlane: Single);
begin
  // Set attributes for the projection matrix
  m_fFOV        := fFOV;
  m_fAspect     := fAspect;
  m_fNearPlane  := fNearPlane;
  m_fFarPlane   := fFarPlane;

  D3DXMatrixPerspectiveFovLH(m_mProj, fFOV, fAspect, fNearPlane, fFarPlane);
end;


procedure CBaseCamera.SetInvertPitch(bInvertPitch: Boolean);
begin
  m_bInvertPitch := bInvertPitch;
end;

procedure CBaseCamera.SetDrag(bMovementDrag: Boolean; fTotalDragTimeToZero: Single = 0.25);
begin
  m_bMovementDrag := bMovementDrag;
  m_fTotalDragTimeToZero := fTotalDragTimeToZero;
end;

procedure CBaseCamera.SetEnableYAxisMovement(bEnableYAxisMovement: Boolean);
begin
  m_bEnableYAxisMovement := bEnableYAxisMovement;
end;

procedure CBaseCamera.SetEnablePositionMovement(bEnablePositionMovement: Boolean);
begin
  m_bEnablePositionMovement := bEnablePositionMovement;
end;

procedure CBaseCamera.SetClipToBoundary(bClipToBoundary: Boolean; const pvMinBoundary, pvMaxBoundary: TD3DXVector3);
begin
  m_bClipToBoundary := bClipToBoundary;
  {if (pvMinBoundary<>nil) then} m_vMinBoundary := pvMinBoundary;
  {if (pvMaxBoundary<>nil) then} m_vMaxBoundary := pvMaxBoundary;
end;

procedure CBaseCamera.SetScalers(fRotationScaler: Single = 0.01; fMoveScaler: Single = 5.0);
begin
  m_fRotationScaler := fRotationScaler;
  m_fMoveScaler := fMoveScaler;
end;

procedure CBaseCamera.SetNumberOfFramesToSmoothMouseData(nFrames: Integer);
begin
  if (nFrames > 0) then m_fFramesToSmoothMouseData := nFrames;
end;

procedure CBaseCamera.SetResetCursorAfterMove(bResetCursorAfterMove: Boolean);
begin
  m_bResetCursorAfterMove := bResetCursorAfterMove;
end;

function CBaseCamera.GetViewMatrix: PD3DXMatrix;
begin
  Result:= @m_mView;
end;

function CBaseCamera.GetProjMatrix: PD3DXMatrix;
begin
  Result:= @m_mProj;
end;

function CBaseCamera.IsBeingDragged: Boolean;
begin
  Result:= (m_bMouseLButtonDown or m_bMouseMButtonDown or m_bMouseRButtonDown);
end;


//-----------------------------------------------------------------------------
// Name: HandleMessages
// Desc: Call this from your message proc so this class can handle window messages
//-----------------------------------------------------------------------------
function CBaseCamera.HandleMessages(hWnd: HWND; uMsg: Cardinal; wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  mappedKey: TD3DUtil_CameraKeys;
begin
  case uMsg of
    WM_KEYDOWN:
    begin
      // Map this key to a D3DUtil_CameraKeys enum and update the
      // state of m_aKeys[] by adding the KEY_WAS_DOWN_MASK|KEY_IS_DOWN_MASK mask
      // only if the key is not down
      mappedKey := MapKey(wParam);
      if (mappedKey <> CAM_UNKNOWN) then
      begin
        if (not IsKeyDown(m_aKeys[mappedKey]))
        then m_aKeys[mappedKey] := KEY_WAS_DOWN_MASK or KEY_IS_DOWN_MASK;
      end;
    end;

    WM_KEYUP:
    begin
      // Map this key to a D3DUtil_CameraKeys enum and update the
      // state of m_aKeys[] by removing the KEY_IS_DOWN_MASK mask.
      mappedKey := MapKey(wParam);
      if (mappedKey <> CAM_UNKNOWN)
      then m_aKeys[mappedKey] := m_aKeys[mappedKey] and not KEY_IS_DOWN_MASK;
    end;

    WM_RBUTTONDOWN,
    WM_MBUTTONDOWN,
    WM_LBUTTONDOWN:
    begin
      // Update member var state
      if (uMsg = WM_LBUTTONDOWN) then begin m_bMouseLButtonDown := True; m_nCurrentButtonMask := m_nCurrentButtonMask or MOUSE_LEFT_BUTTON; end;
      if (uMsg = WM_MBUTTONDOWN) then begin m_bMouseMButtonDown := True; m_nCurrentButtonMask := m_nCurrentButtonMask or MOUSE_MIDDLE_BUTTON; end;
      if (uMsg = WM_RBUTTONDOWN) then begin m_bMouseRButtonDown := True; m_nCurrentButtonMask := m_nCurrentButtonMask or MOUSE_RIGHT_BUTTON; end;

      // Capture the mouse, so if the mouse button is 
      // released outside the window, we'll get the WM_LBUTTONUP message
      SetCapture(hWnd);
      GetCursorPos(m_ptLastMousePosition); 
      Result:= iTrue;
      Exit;
    end;

    WM_RBUTTONUP,
    WM_MBUTTONUP,
    WM_LBUTTONUP:
    begin
      // Update member var state
      if (uMsg = WM_LBUTTONUP) then begin m_bMouseLButtonDown := False; m_nCurrentButtonMask := m_nCurrentButtonMask and not MOUSE_LEFT_BUTTON; end;
      if (uMsg = WM_MBUTTONUP) then begin m_bMouseMButtonDown := False; m_nCurrentButtonMask := m_nCurrentButtonMask and not MOUSE_MIDDLE_BUTTON; end;
      if (uMsg = WM_RBUTTONUP) then begin m_bMouseRButtonDown := False; m_nCurrentButtonMask := m_nCurrentButtonMask and not MOUSE_RIGHT_BUTTON; end;

      // Release the capture if no mouse buttons down
      if (not m_bMouseLButtonDown  and
          not m_bMouseRButtonDown and
          not m_bMouseMButtonDown) then 
      begin
        ReleaseCapture;
      end;
    end;

    WM_MOUSEWHEEL:
      // Update member var state
      m_nMouseWheelDelta := Smallint(HIWORD(wParam)) div 120;
  end;

  Result:= iFalse;
end;


//-----------------------------------------------------------------------------
// Name: UpdateMouseDelta
// Desc: Figure out the mouse delta based on mouse movement
//-----------------------------------------------------------------------------
procedure CBaseCamera.UpdateMouseDelta(fElapsedTime: Single);
var
  ptCurMouseDelta: TPoint;
  ptCurMousePos: TPoint;
  ptCenter: TPoint;
  rcDesktop: TRect;
  fPercentOfNew: Single;
  fPercentOfOld: Single;
begin
  // Get current position of mouse
  GetCursorPos(ptCurMousePos);

  // Calc how far it's moved since last frame
  ptCurMouseDelta.x := ptCurMousePos.x - m_ptLastMousePosition.x;
  ptCurMouseDelta.y := ptCurMousePos.y - m_ptLastMousePosition.y;

  // Record current position for next time
  m_ptLastMousePosition := ptCurMousePos;

  if m_bResetCursorAfterMove then
  begin
    // Set position of camera to center of desktop,
    // so it always has room to move.  This is very useful
    // if the cursor is hidden.  If this isn't done and cursor is hidden,
    // then invisible cursor will hit the edge of the screen
    // and the user can't tell what happened
    GetWindowRect(GetDesktopWindow, rcDesktop);
    ptCenter.x := (rcDesktop.right - rcDesktop.left) div 2;
    ptCenter.y := (rcDesktop.bottom - rcDesktop.top) div 2;
    SetCursorPos(ptCenter.x, ptCenter.y);
    m_ptLastMousePosition := ptCenter;
  end;

  // Smooth the relative mouse data over a few frames so it isn't
  // jerky when moving slowly at low frame rates.
  fPercentOfNew :=  1.0 / m_fFramesToSmoothMouseData;
  fPercentOfOld :=  1.0 - fPercentOfNew;
  m_vMouseDelta.x := m_vMouseDelta.x*fPercentOfOld + ptCurMouseDelta.x*fPercentOfNew;
  m_vMouseDelta.y := m_vMouseDelta.y*fPercentOfOld + ptCurMouseDelta.y*fPercentOfNew;

  D3DXVec2Scale(m_vRotVelocity, m_vMouseDelta, m_fRotationScaler)
end;


//-----------------------------------------------------------------------------
// Name: UpdateVelocity
// Desc: Figure out the velocity based on keyboard input & drag if any
//-----------------------------------------------------------------------------
procedure CBaseCamera.UpdateVelocity(fElapsedTime: Single);
var
//  mRotDelta: TD3DXMatrix; //BAA: It's never used anyway
  vAccel: TD3DXVector3;
  v: TD3DXVector3;
begin
  vAccel := D3DXVector3Zero;

  if m_bEnablePositionMovement then
  begin
    // Update acceleration vector based on keyboard state
    if IsKeyDown(m_aKeys[CAM_MOVE_FORWARD ]) then vAccel.z := vAccel.z + 1.0;
    if IsKeyDown(m_aKeys[CAM_MOVE_BACKWARD]) then vAccel.z := vAccel.z - 1.0;
    if (m_bEnableYAxisMovement) then
    begin
      if IsKeyDown(m_aKeys[CAM_MOVE_UP  ]) then vAccel.y := vAccel.y + 1.0;
      if IsKeyDown(m_aKeys[CAM_MOVE_DOWN]) then vAccel.y := vAccel.y - 1.0;
    end;
    if IsKeyDown(m_aKeys[CAM_STRAFE_RIGHT]) then vAccel.x := vAccel.x + 1.0;
    if IsKeyDown(m_aKeys[CAM_STRAFE_LEFT ]) then vAccel.x := vAccel.x - 1.0;
  end;

  // Normalize vector so if moving 2 dirs (left & forward),
  // the camera doesn't move faster than if moving in 1 dir
  D3DXVec3Normalize(vAccel, vAccel);

  // Scale the acceleration vector
  D3DXVec3Scale(vAccel, vAccel, m_fMoveScaler);

  if m_bMovementDrag then
  begin
    // Is there any acceleration this frame?
    if (D3DXVec3LengthSq(vAccel ) > 0) then
    begin
      // If so, then this means the user has pressed a movement key\
      // so change the velocity immediately to acceleration
      // upon keyboard input.  This isn't normal physics
      // but it will give a quick response to keyboard input
      m_vVelocity := vAccel;
      m_fDragTimer := m_fTotalDragTimeToZero;
      D3DXVec3Scale(m_vVelocityDrag, vAccel, 1 / m_fDragTimer);
    end else
    begin
      // If no key being pressed, then slowly decrease velocity to 0
      if (m_fDragTimer > 0) then
      begin
        // Drag until timer is <= 0
        // m_vVelocity := m_vVelocity - m_vVelocityDrag * fElapsedTime;
        D3DXVec3Subtract(m_vVelocity, m_vVelocity, D3DXVec3Scale(v, m_vVelocityDrag, fElapsedTime)^);
        m_fDragTimer := m_fDragTimer - fElapsedTime;
      end else
      begin
        // Zero velocity
        m_vVelocity := D3DXVector3Zero;
      end;
    end;
  end else
  begin
    // No drag, so immediatly change the velocity
    m_vVelocity := vAccel;
  end;
end;


function CBaseCamera.IsKeyDown(key: Byte): Boolean;
begin
  Result:= ((key and KEY_IS_DOWN_MASK) = KEY_IS_DOWN_MASK);
end;


function CBaseCamera.WasKeyDown(key: Byte): Boolean;
begin
  Result:= ((key and KEY_WAS_DOWN_MASK) = KEY_WAS_DOWN_MASK);
end;


//-----------------------------------------------------------------------------
// Name: ConstrainToBoundary
// Desc: Clamps pV to lie inside m_vMinBoundary & m_vMaxBoundary
//-----------------------------------------------------------------------------
procedure CBaseCamera.ConstrainToBoundary(var pV: TD3DXVector3);
begin
  // Constrain vector to a bounding box
  pV.x := Max(pV.x, m_vMinBoundary.x);
  pV.y := Max(pV.y, m_vMinBoundary.y);
  pV.z := Max(pV.z, m_vMinBoundary.z);

  pV.x := Min(pV.x, m_vMaxBoundary.x);
  pV.y := Min(pV.y, m_vMaxBoundary.y);
  pV.z := Min(pV.z, m_vMaxBoundary.z);
end;




//-----------------------------------------------------------------------------
// Name: MapKey
// Desc: Maps a windows virtual key to an enum
//-----------------------------------------------------------------------------
function CBaseCamera.MapKey(nKey: LongWord): TD3DUtil_CameraKeys;
begin
  // This could be upgraded to a method that's user-definable but for
  // simplisity, we'll use a hardcoded mapping.
  case nKey of
    VK_LEFT:  Result:= CAM_STRAFE_LEFT;
    VK_RIGHT: Result:= CAM_STRAFE_RIGHT;
    VK_UP:    Result:= CAM_MOVE_FORWARD;
    VK_DOWN:  Result:= CAM_MOVE_BACKWARD;
    VK_PRIOR: Result:= CAM_MOVE_UP;        // pgup
    VK_NEXT:  Result:= CAM_MOVE_DOWN;      // pgdn

    Ord('A'):   Result:= CAM_STRAFE_LEFT;
    Ord('D'):   Result:= CAM_STRAFE_RIGHT;
    Ord('W'):   Result:= CAM_MOVE_FORWARD;
    Ord('S'):   Result:= CAM_MOVE_BACKWARD;
    Ord('Q'):   Result:= CAM_MOVE_DOWN;
    Ord('E'):   Result:= CAM_MOVE_UP;

    VK_NUMPAD4: Result:= CAM_STRAFE_LEFT;
    VK_NUMPAD6: Result:= CAM_STRAFE_RIGHT;
    VK_NUMPAD8: Result:= CAM_MOVE_FORWARD;
    VK_NUMPAD2: Result:= CAM_MOVE_BACKWARD;
    VK_NUMPAD9: Result:= CAM_MOVE_UP;
    VK_NUMPAD3: Result:= CAM_MOVE_DOWN;

    VK_HOME:   Result:= CAM_RESET;
  else
    Result:= CAM_UNKNOWN;
  end;
end;




//-----------------------------------------------------------------------------
// Name: Reset
// Desc: Reset the camera's position back to the default
//-----------------------------------------------------------------------------
procedure CBaseCamera.Reset;
begin
  SetViewParams(m_vDefaultEye, m_vDefaultLookAt);
end;




//-----------------------------------------------------------------------------
// Name: CFirstPersonCamera
// Desc: Constructor
//-----------------------------------------------------------------------------
constructor CFirstPersonCamera.Create;
begin
  inherited Create;
end;




//-----------------------------------------------------------------------------
// Name: FrameMove
// Desc: Update the view matrix based on user input & elapsed time
//-----------------------------------------------------------------------------
procedure CFirstPersonCamera.FrameMove(fElapsedTime: Single);
var
  vPosDelta: TD3DXVector3;
  fYawDelta: Single;
  fPitchDelta: Single;
  mCameraRot: TD3DXMatrix;
  vWorldUp, vWorldAhead: TD3DXVector3;
  vLocalUp: TD3DXVector3;
  vLocalAhead: TD3DXVector3;
  vPosDeltaWorld: TD3DXVector3;
const D3DX_PI = 3.14;
begin
  if IsKeyDown(m_aKeys[CAM_RESET]) then Reset;

  // Get the mouse movement (if any) if the mouse button are down
  if (m_bMouseLButtonDown or m_bMouseMButtonDown or m_bMouseRButtonDown)
  then UpdateMouseDelta(fElapsedTime);

  // Get amount of velocity based on the keyboard input and drag (if any)
  UpdateVelocity(fElapsedTime);

  // Simple euler method to calculate position delta
  D3DXVec3Scale(vPosDelta, m_vVelocity, fElapsedTime);

  // If rotating the camera
  if (m_bMouseLButtonDown or m_bMouseMButtonDown or m_bMouseRButtonDown) then
  begin
    // Update the pitch & yaw angle based on mouse movement
    fYawDelta   := m_vRotVelocity.x;
    fPitchDelta := m_vRotVelocity.y;

    // Invert pitch if requested
    if (m_bInvertPitch) then fPitchDelta := - fPitchDelta;

    m_fCameraPitchAngle := m_fCameraPitchAngle + fPitchDelta;
    m_fCameraYawAngle   := m_fCameraYawAngle + fYawDelta;

    // Limit pitch to straight up or straight down
{$IFDEF FPC}
    //todo: FIX ME! FPC compiler bug!
    if (-D3DX_PI/2.0) > m_fCameraPitchAngle then m_fCameraPitchAngle := (-D3DX_PI/2.0);
    if (+D3DX_PI/2.0) < m_fCameraPitchAngle then m_fCameraPitchAngle := (+D3DX_PI/2.0);
{$ELSE}
    m_fCameraPitchAngle := Max(-D3DX_PI/2.0,  m_fCameraPitchAngle);
    m_fCameraPitchAngle := Min(+D3DX_PI/2.0,  m_fCameraPitchAngle);
{$ENDIF}    
  end;

  // Make a rotation matrix based on the camera's yaw & pitch
  D3DXMatrixRotationYawPitchRoll(mCameraRot, m_fCameraYawAngle, m_fCameraPitchAngle, 0);

  // Transform vectors based on camera's rotation matrix
  vLocalUp    := D3DXVector3(0,1,0);
  vLocalAhead := D3DXVector3(0,0,1);
  D3DXVec3TransformCoord(vWorldUp, vLocalUp, mCameraRot);
  D3DXVec3TransformCoord(vWorldAhead, vLocalAhead, mCameraRot);

  // Transform the position delta by the camera's rotation
  D3DXVec3TransformCoord(vPosDeltaWorld, vPosDelta, mCameraRot);
  if (not m_bEnableYAxisMovement) then vPosDeltaWorld.y := 0.0;

  // Move the eye position
  D3DXVec3Add(m_vEye, m_vEye, vPosDeltaWorld);
  if m_bClipToBoundary then ConstrainToBoundary(m_vEye);

  // Update the lookAt position based on the eye position 
  D3DXVec3Add(m_vLookAt, m_vEye, vWorldAhead);

  // Update the view matrix
  D3DXMatrixLookAtLH(m_mView, m_vEye, m_vLookAt, vWorldUp);

  D3DXMatrixInverse(m_mCameraWorld, nil, m_mView);
end;


function CFirstPersonCamera.GetWorldRight: TD3DXVector3;
begin
  Result:= D3DXVector3(m_mCameraWorld._11, m_mCameraWorld._12, m_mCameraWorld._13);
end;

function CFirstPersonCamera.GetWorldUp: TD3DXVector3;
begin
  Result:= D3DXVector3(m_mCameraWorld._21, m_mCameraWorld._22, m_mCameraWorld._23);
end;

function CFirstPersonCamera.GetWorldAhead: TD3DXVector3;
begin
  Result:= D3DXVector3(m_mCameraWorld._31, m_mCameraWorld._32, m_mCameraWorld._33);
end;

function CFirstPersonCamera.GetEyePt: TD3DXVector3;
begin
  Result:= D3DXVector3(m_mCameraWorld._41, m_mCameraWorld._42, m_mCameraWorld._43);
end;


//-----------------------------------------------------------------------------
// Name: CModelViewerCamera
// Desc: Constructor 
//-----------------------------------------------------------------------------
constructor CModelViewerCamera.Create;
begin
  inherited Create;

  m_WorldArcBall:= CD3DArcBall.Create;
  m_ViewArcBall:= CD3DArcBall.Create;

  D3DXMatrixIdentity(m_mWorld);
  D3DXMatrixIdentity(m_mModelRot);
  D3DXMatrixIdentity(m_mModelLastRot);
  m_vModelCenter := D3DXVector3Zero;
  m_fRadius    := 5.0;
  m_fDefaultRadius := 5.0;
  m_fMinRadius := 1.0;
  m_fMaxRadius := D3DX_DEFAULT; // FLT_MAX;
  m_bLimitPitch := False;
  m_bEnablePositionMovement := False;

  m_nRotateModelButtonMask  := MOUSE_LEFT_BUTTON;
  m_nZoomButtonMask         := MOUSE_WHEEL;
  m_nRotateCameraButtonMask := MOUSE_RIGHT_BUTTON;
end;

destructor CModelViewerCamera.Destroy;
begin
  m_WorldArcBall.Free;
  m_ViewArcBall.Free;
  inherited Destroy;
end;



//-----------------------------------------------------------------------------
// Name: FrameMove
// Desc: Update the view matrix & the model's world matrix based
//       on user input & elapsed time
//-----------------------------------------------------------------------------
procedure CModelViewerCamera.FrameMove(fElapsedTime: Single);
var
  vPosDelta: TD3DXVector3;
  mCameraRot: TD3DXMatrix;
  vWorldUp, vWorldAhead: TD3DXVector3;
  vLocalUp: TD3DXVector3;
  vLocalAhead: TD3DXVector3;
  vPosDeltaWorld: TD3DXVector3;
  v: TD3DXVector3;
  mInvView: TD3DXMatrix;
  mModelLastRotInv: TD3DXMatrix;
  mModelRot: TD3DXMatrix;
  m: TD3DXMatrix;
  pXBasis: PD3DXVector3;
  pYBasis: PD3DXVector3;
  pZBasis: PD3DXVector3;
  mTrans: TD3DXMatrix;
begin
  if IsKeyDown(m_aKeys[CAM_RESET]) then Reset;

  // Get the mouse movement (if any) if the mouse button are down
  if (m_nCurrentButtonMask <> 0) then UpdateMouseDelta(fElapsedTime);

  // Get amount of velocity based on the keyboard input and drag (if any)
  UpdateVelocity(fElapsedTime);

  // Simple euler method to calculate position delta
  D3DXVec3Scale(vPosDelta, m_vVelocity, fElapsedTime);

  // Change the radius from the camera to the model based on wheel scrolling
  if (m_nMouseWheelDelta <> 0) and (m_nZoomButtonMask = MOUSE_WHEEL) then
    m_fRadius := m_fRadius - m_nMouseWheelDelta * m_fRadius * 0.1;
  m_fRadius := Min(m_fMaxRadius, m_fRadius);
  m_fRadius := Max(m_fMinRadius, m_fRadius);
  m_nMouseWheelDelta := 0;

  // Get the inverse of the arcball's rotation matrix
  D3DXMatrixInverse(mCameraRot, nil, m_ViewArcBall.GetRotationMatrix^);

  // Transform vectors based on camera's rotation matrix
  vLocalUp    := D3DXVector3(0,1,0);
  vLocalAhead := D3DXVector3(0,0,1);
  D3DXVec3TransformCoord(vWorldUp, vLocalUp, mCameraRot);
  D3DXVec3TransformCoord(vWorldAhead, vLocalAhead, mCameraRot);

  // Transform the position delta by the camera's rotation
  D3DXVec3TransformCoord(vPosDeltaWorld, vPosDelta, mCameraRot);

  // Move the lookAt position
  D3DXVec3Add(m_vLookAt, m_vLookAt, vPosDeltaWorld);
  if m_bClipToBoundary then ConstrainToBoundary(m_vLookAt);

  // Update the eye point based on a radius away from the lookAt position
  // m_vEye = m_vLookAt - vWorldAhead * m_fRadius;
  D3DXVec3Subtract(m_vEye, m_vLookAt, D3DXVec3Scale(v, vWorldAhead, m_fRadius)^);

  // Update the view matrix
  D3DXMatrixLookAtLH(m_mView, m_vEye, m_vLookAt, vWorldUp);

  D3DXMatrixInverse(mInvView, nil, m_mView);
  mInvView._41 := 0; mInvView._42 := 0; mInvView._43 := 0;

  D3DXMatrixInverse(mModelLastRotInv, nil, m_mModelLastRot);

  // Accumulate the delta of the arcball's rotation in view space.
  // Note that per-frame delta rotations could be problematic over long periods of time.
  mModelRot := m_WorldArcBall.GetRotationMatrix^;
  // m_mModelRot *= m_mView * mModelLastRotInv * mModelRot * mInvView;
  D3DXMatrixMultiply(m, mModelRot, mInvView);
  D3DXMatrixMultiply(m, mModelLastRotInv, m);
  D3DXMatrixMultiply(m, m_mView, m);
  D3DXMatrixMultiply(m_mModelRot, m_mModelRot, m);

  m_mModelLastRot := mModelRot;

  // Since we're accumulating delta rotations, we need to orthonormalize
  // the matrix to prevent eventual matrix skew
  pXBasis := PD3DXVector3(@m_mWorld._11);
  pYBasis := PD3DXVector3(@m_mWorld._21);
  pZBasis := PD3DXVector3(@m_mWorld._31);
  D3DXVec3Normalize(pXBasis^, pXBasis^);
  D3DXVec3Cross(pYBasis^, pZBasis^, pXBasis^);
  D3DXVec3Normalize(pYBasis^, pYBasis^);
  D3DXVec3Cross(pZBasis^, pXBasis^, pYBasis^);

  // Translate the rotation matrix to the same position as the lookAt position
  m_mModelRot._41 := m_vLookAt.x;
  m_mModelRot._42 := m_vLookAt.y;
  m_mModelRot._43 := m_vLookAt.z;

  // Translate world matrix so its at the center of the model
  D3DXMatrixTranslation(mTrans, -m_vModelCenter.x, -m_vModelCenter.y, -m_vModelCenter.z);
  D3DXMatrixMultiply(m_mWorld, mTrans, m_mModelRot);
end;




//-----------------------------------------------------------------------------
// Name: Reset
// Desc: Reset the camera's position back to the default
//-----------------------------------------------------------------------------
procedure CModelViewerCamera.Reset;
begin
  inherited Reset;

  D3DXMatrixIdentity(m_mWorld);
  m_fRadius := m_fDefaultRadius;
  m_WorldArcBall.Reset;
  m_ViewArcBall.Reset;
end;



//-----------------------------------------------------------------------------
// Name: HandleMessages
// Desc: Call this from your message proc so this class can handle window messages
//-----------------------------------------------------------------------------
function CModelViewerCamera.HandleMessages(hWnd: HWND; uMsg: Cardinal; wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  iMouseX: Integer;
  iMouseY: Integer;
begin
  inherited HandleMessages(hWnd, uMsg, wParam, lParam);

  if ((uMsg = WM_LBUTTONDOWN) and (m_nRotateModelButtonMask = MOUSE_LEFT_BUTTON)) or
     ((uMsg = WM_MBUTTONDOWN) and (m_nRotateModelButtonMask = MOUSE_MIDDLE_BUTTON)) or
     ((uMsg = WM_RBUTTONDOWN) and (m_nRotateModelButtonMask = MOUSE_RIGHT_BUTTON)) then
  begin
    iMouseX := LOWORD(lParam); // GET_X_LPARAM(lParam);
    iMouseY := HIWORD(lParam); // GET_Y_LPARAM(lParam);
    m_WorldArcBall.OnBegin(iMouseX, iMouseY);
  end;

  if ((uMsg = WM_LBUTTONDOWN) and (m_nRotateCameraButtonMask = MOUSE_LEFT_BUTTON)) or
     ((uMsg = WM_MBUTTONDOWN) and (m_nRotateCameraButtonMask = MOUSE_MIDDLE_BUTTON)) or
     ((uMsg = WM_RBUTTONDOWN) and (m_nRotateCameraButtonMask = MOUSE_RIGHT_BUTTON)) then
  begin
    iMouseX := LOWORD(lParam); // GET_X_LPARAM(lParam);
    iMouseY := HIWORD(lParam); // GET_Y_LPARAM(lParam);
    m_ViewArcBall.OnBegin(iMouseX, iMouseY);
  end;

  if (uMsg = WM_MOUSEMOVE) then
  begin
    iMouseX := LOWORD(lParam); // GET_X_LPARAM(lParam);
    iMouseY := HIWORD(lParam); // GET_Y_LPARAM(lParam);
    m_WorldArcBall.OnMove(iMouseX, iMouseY);
    m_ViewArcBall.OnMove(iMouseX, iMouseY);
  end;

  if ((uMsg = WM_LBUTTONUP) and (m_nRotateModelButtonMask = MOUSE_LEFT_BUTTON)) or
     ((uMsg = WM_MBUTTONUP) and (m_nRotateModelButtonMask = MOUSE_MIDDLE_BUTTON)) or
     ((uMsg = WM_RBUTTONUP) and (m_nRotateModelButtonMask = MOUSE_RIGHT_BUTTON)) then
  begin
    m_WorldArcBall.OnEnd;
  end;

  if ((uMsg = WM_LBUTTONUP) and (m_nRotateCameraButtonMask = MOUSE_LEFT_BUTTON)) or
     ((uMsg = WM_MBUTTONUP) and (m_nRotateCameraButtonMask = MOUSE_MIDDLE_BUTTON)) or
     ((uMsg = WM_RBUTTONUP) and (m_nRotateCameraButtonMask = MOUSE_RIGHT_BUTTON)) then
  begin
    m_ViewArcBall.OnEnd;
  end;

  Result:= iFalse;
end;

procedure CModelViewerCamera.SetButtonMasks(nRotateModelButtonMask: Integer = MOUSE_LEFT_BUTTON; nZoomButtonMask: Integer = MOUSE_WHEEL; nRotateCameraButtonMask: Integer = MOUSE_RIGHT_BUTTON); { m_nRotateModelButtonMask = nRotateModelButtonMask, m_nZoomButtonMask = nZoomButtonMask; m_nRotateCameraButtonMask = nRotateCameraButtonMask; }
begin
  m_nRotateModelButtonMask := nRotateModelButtonMask;
  m_nZoomButtonMask := nZoomButtonMask;
  m_nRotateCameraButtonMask := nRotateCameraButtonMask;
end;

procedure CModelViewerCamera.SetWindow(nWidth, nHeight: Integer; fArcballRadius: Single = 0.9); { m_WorldArcBall.SetWindow( nWidth, nHeight, fArcballRadius ); m_ViewArcBall.SetWindow( nWidth, nHeight, fArcballRadius ); }
begin
  m_WorldArcBall.SetWindow(nWidth, nHeight, fArcballRadius);
  m_ViewArcBall.SetWindow (nWidth, nHeight, fArcballRadius);
end;

procedure CModelViewerCamera.SetRadius(fDefaultRadius: Single = 5.0; fMinRadius: Single = 1.0; fMaxRadius: Single = 3.402823466e+38{FLT_MAX}); { m_fDefaultRadius = m_fRadius = fDefaultRadius; m_fMinRadius = fMinRadius; m_fMaxRadius = fMaxRadius; }
begin
  m_fDefaultRadius := fDefaultRadius; m_fRadius := fDefaultRadius;
  m_fMinRadius := fMinRadius;
  m_fMaxRadius := fMaxRadius;
end;

procedure CModelViewerCamera.SetModelCenter(vModelCenter: TD3DXVector3);
begin
  m_vModelCenter := vModelCenter;
end;

procedure CModelViewerCamera.SetLimitPitch(bLimitPitch: Boolean);
begin
  m_bLimitPitch := bLimitPitch;
end;

procedure CModelViewerCamera.SetViewQuat(q: TD3DXQuaternion);
begin
  m_ViewArcBall.SetQuatNow(q);
end;

procedure CModelViewerCamera.SetWorldQuat(q: TD3DXQuaternion);
begin
  m_WorldArcBall.SetQuatNow(q);
end;

function CModelViewerCamera.GetWorldMatrix: PD3DXMatrix;
begin
  Result:= @m_mWorld;
end;

end.

