{******************************************************************************}
{*                                                                            *}
{*  Copyright (c) 2002, NVIDIA Corporation.                                   *}
{*                                                                            *}
{*  Files:    cgD3D.h, cgContextManager.h, cgProgramManager.h                 *}
{*  Content:  NVIDIA Cg Direct3D interface include files                      *}
{*                                                                            *}
{*  NVIDIA "Cg" Release 1.1 ObjectPascal adaptation by Alexey Barkovoy        *}
{*  E-Mail: clootie@reactor.ru                                                *}
{*                                                                            *}
{*  Modified: 21-Mar-2003                                                     *}
{*                                                                            *}
{*  Latest version can be downloaded from:                                    *}
{*     http://clootie.narod.ru                                                *}
{*                                                                            *}
{******************************************************************************}
{                                                                              }
{ Obtained through: Joint Endeavour of Delphi Innovators (Project JEDI)        }
{                                                                              }
{ The contents of this file are used with permission, subject to the Mozilla   }
{ Public License Version 1.1 (the "License"); you may not use this file except }
{ in compliance with the License. You may obtain a copy of the License at      }
{ http://www.mozilla.org/MPL/MPL-1.1.html                                      }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}

(*
 *
 * Copyright (c) 2002, NVIDIA Corporation.
 *
 *
 *
 * NVIDIA Corporation("NVIDIA") supplies this software to you in consideration
 * of your agreement to the following terms, and your use, installation,
 * modification or redistribution of this NVIDIA software constitutes
 * acceptance of these terms.  If you do not agree with these terms, please do
 * not use, install, modify or redistribute this NVIDIA software.
 *
 *
 * 
 * In consideration of your agreement to abide by the following terms, and 
 * subject to these terms, NVIDIA grants you a personal, non-exclusive license,
 * under NVIDIAs copyrights in this original NVIDIA software (the "NVIDIA 
 * Software"), to use, reproduce, modify and redistribute the NVIDIA 
 * Software, with or without modifications, in source and/or binary forms; 
 * provided that if you redistribute the NVIDIA Software, you must retain the 
 * copyright notice of NVIDIA, this notice and the following text and
 * disclaimers in all such redistributions of the NVIDIA Software. Neither the 
 * name, trademarks, service marks nor logos of NVIDIA Corporation may be used 
 * to endorse or promote products derived from the NVIDIA Software without
 * specific prior written permission from NVIDIA.  Except as expressly stated
 * in this notice, no other rights or licenses express or implied, are granted 
 * by NVIDIA herein, including but not limited to any patent rights that may be 
 * infringed by your derivative works or by other works in which the NVIDIA 
 * Software may be incorporated. No hardware is licensed hereunder. 
 * 
 *
 * 
 * THE NVIDIA SOFTWARE IS BEING PROVIDED ON AN "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING 
 * WITHOUT LIMITATION, WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT, 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR ITS USE AND OPERATION 
 * EITHER ALONE OR IN COMBINATION WITH OTHER PRODUCTS.
 * 
 *  
 * 
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL, 
 * EXEMPLARY, CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOST 
 * PROFITS; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) OR ARISING IN ANY WAY OUT OF THE USE, 
 * REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION OF THE NVIDIA SOFTWARE, 
 * HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT (INCLUDING 
 * NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF NVIDIA HAS BEEN ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

//////////////////////////////////////////////////////////////////////////////
// HISTORY:
// 21-Mar-03 - Alexey Barkovoy:
//   - Updated to Release 1.1 of Cg toolkit (published 04-Mar-2003)
// 11-Jan-03 - Alexey Barkovoy:
//   - Updated to Release 1.0 of Cg toolkit (published 20-Dec-2002)
// 26-Aug-02 - Alexey Barkovoy:
//   - Updated to Beta2 of Cg toolkit (compiler version 1.5)
// 16-Jul-02 - Alexey Barkovoy:
//   - Links to new 1.0.1 DLL version
//   - More robustness & NVIDIA-algo compatible code in objects creation
// 01-Jul-02 - Alexey Barkovoy:
//   - Links to single DLL, uses cg.inc file  
//   - cgD3D_imp.pas file now embedded in implementation part of this unit

unit cgD3D;

interface

uses
  Windows, D3DX8, {$IFDEF DXG_COMPAT}DirectXGraphics{$ELSE}Direct3D8{$ENDIF}, cg;

{$Include JEDI.inc}

const
  CgD3Dlibrary = 'cgD3D_ab.dll';

{$HPPEMIT 'namespace Cgd3d'}
{$HPPEMIT '{'}

type
// to allow multiple streams and user specifed vertex type
// D3DVSDT_FLOAT1      0x00    // 1D float expanded to (value, 0., 0., 1.)
// D3DVSDT_FLOAT2      0x01    // 2D float expanded to (value, value, 0., 1.)
// D3DVSDT_FLOAT3      0x02    // 3D float expanded to (value, value, value, 1.)
// D3DVSDT_FLOAT4      0x03    // 4D float
// D3DVSDT_D3DCOLOR    0x04    // 4D packed unsigned bytes mapped to 0. to 1. range
//                             // Input is in D3DCOLOR format (ARGB) expanded to (R, G, B, A)
// D3DVSDT_UBYTE4      0x05    // 4D unsigned byte
// D3DVSDT_SHORT2      0x06    // 2D signed short expanded to (value, value, 0., 1.)
// D3DVSDT_SHORT4      0x07    // 4D signed short

  // was cgVertexAttribute
  PcgVertexDefinition = ^TcgVertexDefinition;
  TcgVertexDefinition = record
    D3DVSDT_type: Integer; // one of D3DVSDT_*
    name: PChar;           // name of the entry
    stream: Integer;       // stream to put the vertex in
  end;
  cgVertexDefinition = TcgVertexDefinition;
  {$NODEFINE cgVertexDefinition}
  {$NODEFINE TcgVertexDefinition}
  {$HPPEMIT 'typedef cgVertexDefinition TcgVertexDefinition;'}

const
  CGVertexDefinitionEnd: TcgVertexDefinition = (D3DVSDT_type: -1; name: nil; stream: 0);
  {$EXTERNALSYM CGVertexDefinitionEnd}

type
  TcgProgramContainer = class
  private
    pc: Pointer; // handle to VC++ "cgProgramContainer" class
  public
    // get a text definition of what the runtime expects for the vertex
    function GetVertexDeclaration: PChar {const};
    //todo: Check this out
    (*
        Example: returns string containing
        "struct stream0
        {
            D3DXVECTOR3 position;
            D3DXVECTOR2 tex0;
        };"

    *)




    // convert D3DXMATRIX class to type specified in cg program
    // even if your cg program uses 3x3, you can use this to pass a matrix to it
    // upper left hand part of D3DXMATRIX is used to fill the cg matrix
    function SetShaderConstantD3DXMATRIX(iter: PcgBindIter; const m: PD3DXMatrix;
      nArrayElements: Integer{$IFDEF SUPPORTS_DEFAULTPARAMS} = 0{$ENDIF};
      ArrayOffset: Integer{$IFDEF SUPPORTS_DEFAULTPARAMS} = 0{$ENDIF}): HRESULT;



    // All other data formats use this interface
    // The D3D runtime records the size of your data and puts that in the constant fields
    function SetShaderConstant(iter: PcgBindIter; const data;
      nArrayElements: Integer{$IFDEF SUPPORTS_DEFAULTPARAMS} = 0{$ENDIF};
      ArrayOffset: Integer{$IFDEF SUPPORTS_DEFAULTPARAMS} = 0{$ENDIF}): HRESULT;
    // for arrays:
    //   if nArrayElements is specified, write this number of elements rather than the whole array
    //   if ArrayOffset is specified, start writing at this array element
    // otherwise nArrayElements and ArrayOffset are not specified


    // returns texture position 0-8 for the iterator
    // -1 is failure
    function GetTexturePosition(BindIter: PcgBindIter): Integer; overload;
    function GetTexturePosition(const name: PChar): Integer; overload;

    // set the texture via name (needs to search bindings by name, so this is the slowest
    // not recommended
    function SetTexture(const name: PChar; pTexture: IDirect3DBaseTexture8): HRESULT; overload;

    // does a GetTexturePosition, then settexture(n, nTexture)
    function SetTexture(BindIter: PcgBindIter; pTexture: IDirect3DBaseTexture8): HRESULT; overload;

    // does a GetTexturePosition, then SetTextureStageState(n, nTexture, Value)
    function SetTextureStageState(BindIter: PcgBindIter;
      Type_: TD3DTextureStageStateType; Value: DWORD): HRESULT;

    // Direct3D hardcodes wrap state, D3DRS_WRAP1, D3DWRAPCOORD_0 | D3DWRAPCOORD_1
    //   WrapCoords : 0, D3DWRAPCOORD_0, D3DWRAPCOORD_1 or'd together
    function SetTextureWrapMode(BindIter: PcgBindIter; WrapCoords: DWORD): HRESULT;


    // activate this shader (SetVertexShader or SetPixelShader)
    function SetShaderActive: HRESULT;


    // locate a binding within a program by parameter name
    // it is recommended to do this once and use the cgBindIter to reference data
    function GetVertexBindByName(const vertex_attr_name: PChar): PcgBindIter;
    function GetTextureBindByName(const texture_name: PChar): PcgBindIter;
    function GetParameterBindByName(const parameter_name: PChar): PcgBindIter;

    procedure FreeBindIter(iter: PcgBindIter);




    // Get the type for this bind iterator
    function GetBindValueType(const BindIter: PcgBindIter; out array_size: Integer): TcgValueType;
    {
        return cgGetBindValueType(BindIter,  array_size);
    }


    // locate and bindinf iterator by full name
    //  e.g. appdata.texcoord0
    function GetBindByFullName(const parameter_name: PChar): PcgBindIter;

    // from the iterator, get the name of the program
    function GetProgramName: PChar {const};

    function GetProgramObjectCode: PChar {const};

    function GetProgramProfile: TcgProfileType; { return m_ProfileType;}

    // get the shader handle passed to Direct3D
    function GetShaderHandle: DWORD; {return m_dwShaderHandle;}

  public
    constructor Create(handle: Pointer); //Clootie: this is Delphi introduced method
  end;

const
  MAX_STREAMS = 16;

////////////////////////////////////////////////
//
//	cgDX8Vertex  and cgDX8Pixel are defined with
//  macros:
//
//	CG_PROFILE_MACRO(DX8Vertex,dx8vs,"dx8vs")
//	CG_PROFILE_MACRO(DX8Pixel,dx8ps,"dx8ps")
//
//	cgDX8VertexProfile



type
  TcgContextContainer = class
  private
    cc: Pointer; // VC++ object handle
  public
    // retrieve the listing file for the last compile
    function GetLastListing: PChar {const};


    // loads and compiles the cg program
    function LoadCGProgramFromFile(
        const filename: PChar;         // filename of the Cg program
        const title: PChar;            // title of the program
        type_: TcgProfileType;         // profile of the program
        va: PcgVertexDefinition = nil; // optionally, set streams or allow different vetex definitions
        outIndex: PDWORD = nil;
        const entry: PChar = nil): TcgProgramContainer;

    function  LoadCGProgramFromMemory(
        const memory: PChar;           // memory location that contain the cg program
        const title: PChar;            // title of the program
        type_: TcgProfileType;         // profile of the program
        va: PcgVertexDefinition = nil; // optionally, set the streams for each vertex entry
        outIndex: PDWORD = nil;
        const entry: PChar = nil): TcgProgramContainer;


    // use outIndex returned from LoadCGProgramFrom to retriece the program iterator
    function GetProgramIterator(index: DWORD): TcgProgramContainer;


    // program operators (for multiple programs(vertex, pixel) in one cg
    // pass NULL to get first program
    function GetNextProgram(iter: PcgProgramIter): PcgProgramIter;

    // get a specific program
    function ProgramByName(const name: PChar): PcgProgramIter;


    // get the vertexshader or pixel shader handle that is passed to
    // SetVertexShader and SetPixelShader
    function GetShaderHandle(type_: TcgProfileType): DWORD;
    {
        return pm.GetShaderHandle( type);
    }

    constructor Create(handle: Pointer);
    destructor Destroy; override;
  end;


  TcgDirect3D = class
  private
    cd3d: Pointer; // VC++ handle to cgDirect3D class
  public
    constructor Create;
    destructor Destroy; override;

    // add a path to locate the cg programs
    // do not add the trailing '\\'
    procedure AddFilePath(const dirpath: PChar); overload;
    procedure AddFilePath(path: String); overload;


    procedure NotePad(const title: PChar; const listing: PChar);
    {

        std::string text;
        text = title;
        text += listing;

        char * TmpLstFile = tmpnam(0);

        FILE * fpLst = fopen(TmpLstFile, "w");
        fputs(text.c_str(), fpLst);
        fclose(fpLst);

        char Cmd[1000];
        sprintf(Cmd, "notepad %s", TmpLstFile);
        int Ret =  system(Cmd);

        unlink(TmpLstFile);

    }

    function CreateContextContainer(pd3dDevice: IDirect3DDevice8;
      outIndex: PDWORD{$IFDEF SUPPORTS_DEFAULTPARAMS} = nil{$ENDIF}): TcgContextContainer;
    {
        return cm.CreateContextContainer(pd3dDevice,  outIndex);
    }
  end;


const
  CGRTERR_ILLEGAL_PROFILE = -1;
  {$EXTERNALSYM CGRTERR_ILLEGAL_PROFILE}
  CGRTERR_VERTEX_ATTRIBUTES_NOT_SET = -2;
  {$EXTERNALSYM CGRTERR_VERTEX_ATTRIBUTES_NOT_SET}

(*$HPPEMIT '}	/* namespace Cgd3d */'*)

implementation

///////////////////////////////////////////////////////////////////////////////
// Implementation of NVIDIA Cg Direct3D VC classes in DLL as plain C functions

function cgContextContainer_GetLastListing(cc: Pointer): PChar {const}; cdecl; external CgD3Dlibrary;

function cgContextContainer_LoadCGProgramFromFile(cc: Pointer;
    const filename: PChar;           // filename of the Cg program
    const title: PChar;               // title of the program
    type_: TcgProfileType;       // profile of the program
    va: Pointer{PcgVertexAttribute} = nil; // optionally, set streams or allow different vetex definitions
    outIndex: PDWORD = nil;
    const entry: PChar = nil): Pointer{TcgProgramContainer}; cdecl; external CgD3Dlibrary;

function  cgContextContainer_LoadCGProgramFromMemory(cc: Pointer;
    const memory: PChar;          // memory location that contain the cg program
    const title: PChar;               // title of the program
    type_: TcgProfileType;       // profile of the program
    va: Pointer{PcgVertexAttribute} = nil;   // optionally, set the streams for each vertex entry
    outIndex: PDWORD = nil;
    const entry: PChar = nil): Pointer{TcgProgramContainer}; cdecl; external CgD3Dlibrary;

function cgContextContainer_GetProgramIterator(cc: Pointer; index: DWORD): Pointer{TcgProgramContainer}; cdecl; external CgD3Dlibrary;
function cgContextContainer_GetNextProgram(cc: Pointer; iter: PcgProgramIter): PcgProgramIter; cdecl; external CgD3Dlibrary;
function cgContextContainer_ProgramByName(cc: Pointer; const name: PChar): PcgProgramIter; cdecl; external CgD3Dlibrary;
function cgContextContainer_GetShaderHandle(cc: Pointer; type_: TcgProfileType): DWORD; cdecl; external CgD3Dlibrary;
procedure cgContextContainer_Destroy(cc: Pointer); cdecl; external CgD3Dlibrary;

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

function cgDirect3D_Create: Pointer; cdecl; external CgD3Dlibrary;
procedure cgDirect3D_Destroy(cd3d: Pointer); cdecl; external CgD3Dlibrary;
procedure cgDirect3D_AddFilePath(cd3d: Pointer; const dirpath: PChar);  cdecl; external CgD3Dlibrary;
procedure cgDirect3D_NotePad(cd3d: Pointer; const title: PChar; const listing: PChar); cdecl; external CgD3Dlibrary;
function cgDirect3D_CreateContextContainer(cd3d: Pointer; pd3dDevice: IDirect3DDevice8;
  outIndex: PDWORD{$IFDEF SUPPORTS_DEFAULTPARAMS} = nil{$ENDIF}): Pointer{TcgContextContainer}; cdecl; external CgD3Dlibrary;

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

function cgProgramContainer_GetVertexDeclaration(pc: Pointer): PChar {const}; cdecl; external CgD3Dlibrary;

function cgProgramContainer_SetShaderConstantD3DXMATRIX(pc: Pointer; iter: PcgBindIter; const m: PD3DXMatrix;
  nArrayElements: Integer{$IFDEF SUPPORTS_DEFAULTPARAMS} = 0{$ENDIF};
  ArrayOffset: Integer{$IFDEF SUPPORTS_DEFAULTPARAMS} = 0{$ENDIF}): HRESULT; cdecl; external CgD3Dlibrary;

function cgProgramContainer_SetShaderConstant(pc: Pointer; iter: PcgBindIter; const data;
  nArrayElements: Integer{$IFDEF SUPPORTS_DEFAULTPARAMS} = 0{$ENDIF};
  ArrayOffset: Integer{$IFDEF SUPPORTS_DEFAULTPARAMS} = 0{$ENDIF}): HRESULT; cdecl; external CgD3Dlibrary;

function cgProgramContainer_GetTexturePosition(pc: Pointer; BindIter: PcgBindIter): Integer; cdecl; external CgD3Dlibrary;
function cgProgramContainer_GetTexturePositionB(pc: Pointer; const name: PChar): Integer; cdecl; external CgD3Dlibrary;
function cgProgramContainer_SetTexture(pc: Pointer; const name: PChar; pTexture: IDirect3DBaseTexture8): HRESULT; cdecl; external CgD3Dlibrary;
function cgProgramContainer_SetTextureB(pc: Pointer; BindIter: PcgBindIter; pTexture: IDirect3DBaseTexture8): HRESULT; cdecl; external CgD3Dlibrary;

function cgProgramContainer_SetTextureStageState(pc: Pointer; BindIter: PcgBindIter;
  Type_: TD3DTextureStageStateType; Value: DWORD): HRESULT; cdecl; external CgD3Dlibrary;

function cgProgramContainer_SetTextureWrapMode(pc: Pointer; BindIter: PcgBindIter; WrapCoords: DWORD): HRESULT; cdecl; external CgD3Dlibrary;
function cgProgramContainer_SetShaderActive(pc: Pointer): HRESULT; cdecl; external CgD3Dlibrary;

function cgProgramContainer_GetVertexBindByName(pc: Pointer; const vertex_attr_name: PChar): PcgBindIter; cdecl; external CgD3Dlibrary;
function cgProgramContainer_GetTextureBindByName(pc: Pointer; const texture_name: PChar): PcgBindIter; cdecl; external CgD3Dlibrary;
function cgProgramContainer_GetParameterBindByName(pc: Pointer; const parameter_name: PChar): PcgBindIter; cdecl; external CgD3Dlibrary;

procedure cgProgramContainer_FreeBindIter(pc: Pointer; iter: PcgBindIter); cdecl; external CgD3Dlibrary;

function cgProgramContainer_GetBindByFullName(pc: Pointer; const parameter_name: PChar): PcgBindIter; cdecl; external CgD3Dlibrary;
function cgProgramContainer_GetProgramName(pc: Pointer): PChar {const}; cdecl; external CgD3Dlibrary;
function cgProgramContainer_GetProgramObjectCode(pc: Pointer): PChar {const}; cdecl; external CgD3Dlibrary;
function cgProgramContainer_GetProgramProfile(pc: Pointer): TcgProfileType; { return m_ProfileType;} cdecl; external CgD3Dlibrary;
function cgProgramContainer_GetShaderHandle(pc: Pointer): DWORD; {return m_dwShaderHandle;} cdecl; external CgD3Dlibrary;

{ TcgProgramContainer }

constructor TcgProgramContainer.Create(handle: Pointer);
begin
  pc:= handle;
end;

procedure TcgProgramContainer.FreeBindIter(iter: PcgBindIter);
begin
  cgProgramContainer_FreeBindIter(pc, iter);
end;

function TcgProgramContainer.GetBindByFullName(
  const parameter_name: PChar): PcgBindIter;
begin
  Result:= cgProgramContainer_GetBindByFullName(pc, parameter_name);
end;

function TcgProgramContainer.GetBindValueType(const BindIter: PcgBindIter;
  out array_size: Integer): TcgValueType;
begin
  Result:= cgGetBindValueType(BindIter, array_size);
end;

function TcgProgramContainer.GetParameterBindByName(
  const parameter_name: PChar): PcgBindIter;
begin
  Result:= cgProgramContainer_GetParameterBindByName(pc, parameter_name);
end;

function TcgProgramContainer.GetProgramName: PChar;
begin
  Result:= cgProgramContainer_GetProgramName(pc);
end;

function TcgProgramContainer.GetProgramObjectCode: PChar;
begin
  Result:= cgProgramContainer_GetProgramObjectCode(pc);
end;

function TcgProgramContainer.GetProgramProfile: TcgProfileType;
begin
  Result:= cgProgramContainer_GetProgramProfile(pc);
end;

function TcgProgramContainer.GetShaderHandle: DWORD;
begin
  Result:= cgProgramContainer_GetShaderHandle(pc);
end;

function TcgProgramContainer.GetTextureBindByName(
  const texture_name: PChar): PcgBindIter;
begin
  Result:= cgProgramContainer_GetTextureBindByName(pc, texture_name);
end;

function TcgProgramContainer.GetTexturePosition(
  BindIter: PcgBindIter): Integer;
begin
  Result:= cgProgramContainer_GetTexturePosition(pc, BindIter);
end;

function TcgProgramContainer.GetTexturePosition(
  const name: PChar): Integer;
begin
  Result:= cgProgramContainer_GetTexturePositionB(pc, name);
end;

function TcgProgramContainer.GetVertexBindByName(
  const vertex_attr_name: PChar): PcgBindIter;
begin
  Result:= cgProgramContainer_GetVertexBindByName(pc, vertex_attr_name);
end;

function TcgProgramContainer.GetVertexDeclaration: PChar;
begin
  Result:= cgProgramContainer_GetVertexDeclaration(pc);
end;

function TcgProgramContainer.SetShaderActive: HRESULT;
begin
  Result:= cgProgramContainer_SetShaderActive(pc);
end;

function TcgProgramContainer.SetShaderConstant(iter: PcgBindIter;
  const data; nArrayElements, ArrayOffset: Integer): HRESULT;
begin
  Result:= cgProgramContainer_SetShaderConstant(pc, iter, data, nArrayElements, ArrayOffset);
end;

function TcgProgramContainer.SetShaderConstantD3DXMATRIX(iter: PcgBindIter;
  const m: PD3DXMatrix; nArrayElements, ArrayOffset: Integer): HRESULT;
begin
  Result:= cgProgramContainer_SetShaderConstantD3DXMATRIX(pc, iter, m, nArrayElements, ArrayOffset);
end;

function TcgProgramContainer.SetTexture(const name: PChar;
  pTexture: IDirect3DBaseTexture8): HRESULT;
begin
  Result:= cgProgramContainer_SetTexture(pc, name, pTexture);
end;

function TcgProgramContainer.SetTexture(BindIter: PcgBindIter;
  pTexture: IDirect3DBaseTexture8): HRESULT;
begin
  Result:= cgProgramContainer_SetTextureB(pc, BindIter, pTexture);
end;

function TcgProgramContainer.SetTextureStageState(BindIter: PcgBindIter;
  Type_: TD3DTextureStageStateType; Value: DWORD): HRESULT;
begin
  Result:= cgProgramContainer_SetTextureStageState(pc, BindIter, Type_, Value);
end;

function TcgProgramContainer.SetTextureWrapMode(BindIter: PcgBindIter;
  WrapCoords: DWORD): HRESULT;
begin
  Result:= cgProgramContainer_SetTextureWrapMode(pc, BindIter, WrapCoords);
end;


{ TcgContextContainer }

constructor TcgContextContainer.Create(handle: Pointer);
begin
  cc:= handle;
end;

destructor TcgContextContainer.Destroy;
begin
  cgContextContainer_Destroy(cc);
  inherited;
end;

function TcgContextContainer.GetLastListing: PChar;
begin
  Result:= cgContextContainer_GetLastListing(cc);
end;

function TcgContextContainer.GetNextProgram(
  iter: PcgProgramIter): PcgProgramIter;
begin
  Result:= cgContextContainer_GetNextProgram(cc, iter);
end;

function TcgContextContainer.GetProgramIterator(
  index: DWORD): TcgProgramContainer;
begin
  Result:= TcgProgramContainer.Create(cgContextContainer_GetProgramIterator(cc, index));
end;

function TcgContextContainer.GetShaderHandle(type_: TcgProfileType): DWORD;
begin
  Result:= cgContextContainer_GetShaderHandle(cc, type_);
end;

function TcgContextContainer.LoadCGProgramFromFile(const filename,
  title: PChar; type_: TcgProfileType; va: PcgVertexDefinition;
  outIndex: PDWORD; const entry: PChar): TcgProgramContainer;
var
  pc: Pointer;
begin
  pc:= cgContextContainer_LoadCGProgramFromFile(
         cc, filename, title, type_, va, outIndex, entry);

  if (pc = nil) then Result:= nil
  else Result:= TcgProgramContainer.Create(pc);
end;

function TcgContextContainer.LoadCGProgramFromMemory(const memory,
  title: PChar; type_: TcgProfileType; va: PcgVertexDefinition;
  outIndex: PDWORD; const entry: PChar): TcgProgramContainer;
var
  pc: Pointer;
begin
  pc:= cgContextContainer_LoadCGProgramFromMemory(
         cc, memory, title, type_, va, outIndex, entry);

  if (pc = nil) then Result:= nil
  else Result:= TcgProgramContainer.Create(pc);
end;

function TcgContextContainer.ProgramByName(
  const name: PChar): PcgProgramIter;
begin
  Result:= cgContextContainer_ProgramByName(cc, name);
end;

{ TcgDirect3D }

procedure TcgDirect3D.AddFilePath(path: String);
begin
  cgDirect3D_AddFilePath(cd3d, PChar(path));
end;

procedure TcgDirect3D.AddFilePath(const dirpath: PChar);
begin
  cgDirect3D_AddFilePath(cd3d, dirpath);
end;

constructor TcgDirect3D.Create;
begin
  cd3d:= cgDirect3D_Create;
end;

destructor TcgDirect3D.Destroy;
begin
  cgDirect3D_Destroy(cd3d);
  inherited;
end;

function TcgDirect3D.CreateContextContainer(pd3dDevice: IDirect3DDevice8;
  outIndex: PDWORD): TcgContextContainer;
var
  cc: Pointer;
begin
  cc:= cgDirect3D_CreateContextContainer(cd3d, pd3dDevice, outIndex);

  if (cc = nil) then Result:= nil
  else Result:= TcgContextContainer.Create(cc);
end;

procedure TcgDirect3D.NotePad(const title, listing: PChar);
begin
  cgDirect3D_NotePad(cd3d, title, listing);
end;

end.

